from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from arro3.core import ChunkedArray, Table

from lonboard._geoarrow.extension_types import construct_geometry_array

if TYPE_CHECKING:
    import geopandas as gpd


def geopandas_to_geoarrow(
    gdf: gpd.GeoDataFrame,
    columns: list[str] | None = None,
    preserve_index: bool | None = None,  # noqa: FBT001
) -> Table:
    try:
        import pyarrow as pa
    except ImportError as e:
        raise ImportError(
            "pyarrow required for converting GeoPandas to arrow.\n"
            "Run `pip install pyarrow`.",
        ) from e

    df_attr = gdf.drop(columns=[gdf._geometry_column_name])  # noqa: SLF001

    if columns is not None:
        df_attr = df_attr[columns]

    pyarrow_table = pa.Table.from_pandas(df_attr, preserve_index=preserve_index)
    field, geom_arr = construct_geometry_array(
        np.array(gdf.geometry),
        crs=gdf.crs,
    )
    return Table.from_arrow(pyarrow_table).append_column(
        field,
        ChunkedArray([geom_arr]),
    )



from collections.abc import Sequence
from functools import partial
from typing import Any

import traitlets
from ipywidgets import FloatRangeSlider
from ipywidgets.widgets.trait_types import TypedTuple

# Import from source to allow mkdocstrings to link to base class
from ipywidgets.widgets.widget_box import VBox


class MultiRangeSlider(VBox):
    """A widget for multiple ranged sliders.

    This is designed to be used with the
    [DataFilterExtension][lonboard.layer_extension.DataFilterExtension] when you want to
    filter on 2 to 4 columns on the same time.

    If you have only a single filter, use an ipywidgets
    [FloatRangeSlider][ipywidgets.widgets.widget_float.FloatRangeSlider] directly.

    # Example

    ```py
    from ipywidgets import FloatRangeSlider

    slider1 = FloatRangeSlider(
        value=(2, 5),
        min=0,
        max=10,
        step=0.1,
        description="First slider: "
    )
    slider2 = FloatRangeSlider(
        value=(30, 40),
        min=0,
        max=50,
        step=1,
        description="Second slider: "
    )
    multi_slider = MultiRangeSlider([slider1, slider2])
    multi_slider
    ```

    Then to propagate updates to a rendered layer, call
    [jsdlink][ipywidgets.widgets.widget_link.jsdlink] to connect the two widgets.

    ```py
    from ipywidgets import jsdlink

    jsdlink(
        (multi_slider, "value"),
        (layer, "filter_range")
    )
    ```

    As you change the slider, the `filter_range` value on the layer class should be
    updated.
    """

    # We use a tuple to force reassignment to update the list
    # This is because list mutations do not get propagated as events
    # https://github.com/jupyter-widgets/ipywidgets/blob/b2531796d414b0970f18050d6819d932417b9953/python/ipywidgets/ipywidgets/widgets/widget_box.py#L52-L54
    value = TypedTuple(trait=TypedTuple(trait=traitlets.Float())).tag(sync=True)

    def __init__(self, children: Sequence[FloatRangeSlider], **kwargs: Any) -> None:
        """Create a new MultiRangeSlider."""
        if len(children) == 1:
            raise ValueError(
                "Expected more than one slider. "
                "For filtering data from a single column, "
                "use a FloatRangeSlider directly.",
            )

        # We manage a list of lists to match what deck.gl expects for the
        # DataFilterExtension
        def callback(change: dict, *, i: int) -> None:
            value = list(self.value)
            value[i] = change["new"]
            self.set_trait("value", value)
            self.send_state("value")

        initial_values = []
        for i, child in enumerate(children):
            func = partial(callback, i=i)
            child.observe(func, "value")
            initial_values.append(child.value)

        super().__init__(children, value=initial_values, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

from lonboard._constants import EXTENSION_NAME

if TYPE_CHECKING:
    from arro3.core import Field


def is_primitive_geoarrow(field: Field) -> bool:
    """Return True if this GeoArrow column is in a "primitive" coordinate representation.

    "Primitive" means both that it is "native" (i.e. not WKB or WKT) and that it
    is a single geometry type (e.g. Point, LineString, Polygon, etc.).

    This will return false for WKB and WKT-serialized arrays. This will also return
    False for Geometry and GeometryCollection type arrays.
    """
    return field.metadata.get(b"ARROW:extension:name") in {
        EXTENSION_NAME.POINT,
        EXTENSION_NAME.LINESTRING,
        EXTENSION_NAME.POLYGON,
        EXTENSION_NAME.MULTIPOINT,
        EXTENSION_NAME.MULTILINESTRING,
        EXTENSION_NAME.MULTIPOLYGON,
        EXTENSION_NAME.BOX,
    }



# Vendored from matplotlib at
# https://github.com/matplotlib/matplotlib/blob/6166e5b990b32c7f46847782c0354653e483560d/lib/matplotlib/colors.py#L75
# ruff: noqa

from __future__ import annotations

import re
from numbers import Real
from typing import Tuple

import numpy as np

from ._color_data import BASE_COLORS, CSS4_COLORS, TABLEAU_COLORS, XKCD_COLORS


class _ColorMapping(dict):
    def __init__(self, mapping):
        super().__init__(mapping)
        self.cache = {}

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.cache.clear()

    def __delitem__(self, key):
        super().__delitem__(key)
        self.cache.clear()


_colors_full_map = {}
# Set by reverse priority order.
_colors_full_map.update(XKCD_COLORS)
_colors_full_map.update(
    {k.replace("grey", "gray"): v for k, v in XKCD_COLORS.items() if "grey" in k},
)
_colors_full_map.update(CSS4_COLORS)
_colors_full_map.update(TABLEAU_COLORS)
_colors_full_map.update(
    {k.replace("gray", "grey"): v for k, v in TABLEAU_COLORS.items() if "gray" in k},
)
_colors_full_map.update(BASE_COLORS)
_colors_full_map = _ColorMapping(_colors_full_map)


def _to_rgba_no_colorcycle(c, alpha: float | None = None) -> Tuple[float, ...]:
    """Convert *c* to an RGBA color, with no support for color-cycle syntax.

    If *alpha* is given, force the alpha value of the returned RGBA tuple
    to *alpha*. Otherwise, the alpha value from *c* is used, if it has alpha
    information, or defaults to 1.

    *alpha* is ignored for the color value ``"none"`` (case-insensitive),
    which always maps to ``(0, 0, 0, 0)``.
    """
    if alpha is not None and not 0 <= alpha <= 1:
        raise ValueError("'alpha' must be between 0 and 1, inclusive")
    orig_c = c
    if c is np.ma.masked:
        return (0.0, 0.0, 0.0, 0.0)
    if isinstance(c, str):
        if c.lower() == "none":
            return (0.0, 0.0, 0.0, 0.0)
        # Named color.
        try:
            # This may turn c into a non-string, so we check again below.
            c = _colors_full_map[c]
        except KeyError:
            if len(orig_c) != 1:
                try:
                    c = _colors_full_map[c.lower()]
                except KeyError:
                    pass
    if isinstance(c, str):
        # hex color in #rrggbb format.
        match = re.match(r"\A#[a-fA-F0-9]{6}\Z", c)
        if match:
            return tuple(int(n, 16) / 255 for n in [c[1:3], c[3:5], c[5:7]]) + (
                alpha if alpha is not None else 1.0,
            )
        # hex color in #rgb format, shorthand for #rrggbb.
        match = re.match(r"\A#[a-fA-F0-9]{3}\Z", c)
        if match:
            return tuple(int(n, 16) / 255 for n in [c[1] * 2, c[2] * 2, c[3] * 2]) + (
                alpha if alpha is not None else 1.0,
            )
        # hex color with alpha in #rrggbbaa format.
        match = re.match(r"\A#[a-fA-F0-9]{8}\Z", c)
        if match:
            color = [int(n, 16) / 255 for n in [c[1:3], c[3:5], c[5:7], c[7:9]]]
            if alpha is not None:
                color[-1] = alpha
            return tuple(color)
        # hex color with alpha in #rgba format, shorthand for #rrggbbaa.
        match = re.match(r"\A#[a-fA-F0-9]{4}\Z", c)
        if match:
            color = [int(n, 16) / 255 for n in [c[1] * 2, c[2] * 2, c[3] * 2, c[4] * 2]]
            if alpha is not None:
                color[-1] = alpha
            return tuple(color)
        # string gray.
        try:
            c = float(c)
        except ValueError:
            pass
        else:
            if not (0 <= c <= 1):
                raise ValueError(
                    f"Invalid string grayscale value {orig_c!r}. "
                    f"Value must be within 0-1 range",
                )
            return c, c, c, alpha if alpha is not None else 1.0
        raise ValueError(f"Invalid RGBA argument: {orig_c!r}")
    # turn 2-D array into 1-D array
    if isinstance(c, np.ndarray):
        if c.ndim == 2 and c.shape[0] == 1:
            c = c.reshape(-1)
    # tuple color.
    if not np.iterable(c):
        raise ValueError(f"Invalid RGBA argument: {orig_c!r}")
    if len(c) not in [3, 4]:  # type: ignore
        raise ValueError("RGBA sequence should have length 3 or 4")
    if not all(isinstance(x, Real) for x in c):
        # Checks that don't work: `map(float, ...)`, `np.array(..., float)` and
        # `np.array(...).astype(float)` would all convert "0.5" to 0.5.
        raise ValueError(f"Invalid RGBA argument: {orig_c!r}")
    # Return a tuple to prevent the cached value from being modified.
    c = tuple(map(float, c))
    if len(c) == 3 and alpha is None:
        alpha = 1
    if alpha is not None:
        c = c[:3] + (alpha,)
    if any(elem < 0 or elem > 1 for elem in c):
        raise ValueError("RGBA values should be within 0-1 range")
    return c



"""Convert a GeoArrow array to interleaved representation."""

import numpy as np
from arro3.core import (
    Array,
    ChunkedArray,
    DataType,
    Field,
    Table,
    fixed_size_list_array,
    struct_field,
)

from lonboard._constants import EXTENSION_NAME
from lonboard._geoarrow.ops.reproject import (
    _map_coords_nest_0,
    _map_coords_nest_1,
    _map_coords_nest_2,
    _map_coords_nest_3,
)
from lonboard._utils import get_geometry_column_index


def make_geometry_interleaved(
    table: Table,
) -> Table:
    """Convert geometry columns in table to interleaved coordinate layout."""
    geom_col_idx = get_geometry_column_index(table.schema)
    # No geometry column in table
    if geom_col_idx is None:
        return table

    geom_field = table.schema.field(geom_col_idx)
    geom_column = table.column(geom_col_idx)

    # The GeoArrow box extension type is only struct, not interleaved. It will be
    # converted to an interleaved polygon separately, if needed.
    if geom_field.metadata.get(b"ARROW:extension:name") == EXTENSION_NAME.BOX:
        return table

    new_field, new_column = transpose_column(field=geom_field, column=geom_column)
    return table.set_column(geom_col_idx, new_field, new_column)


def transpose_column(
    *,
    field: Field,
    column: ChunkedArray,
) -> tuple[Field, ChunkedArray]:
    extension_type_name = field.metadata[b"ARROW:extension:name"]

    new_chunked_array = _transpose_column(
        column,
        extension_type_name=extension_type_name,  # type: ignore
    )
    return field.with_type(new_chunked_array.type), new_chunked_array


def _transpose_column(
    column: ChunkedArray,
    *,
    extension_type_name: EXTENSION_NAME,
) -> ChunkedArray:
    if extension_type_name == EXTENSION_NAME.POINT:
        func = _transpose_chunk_nest_0
    elif extension_type_name in [EXTENSION_NAME.LINESTRING, EXTENSION_NAME.MULTIPOINT]:
        func = _transpose_chunk_nest_1
    elif extension_type_name in [
        EXTENSION_NAME.POLYGON,
        EXTENSION_NAME.MULTILINESTRING,
    ]:
        func = _transpose_chunk_nest_2

    elif extension_type_name == EXTENSION_NAME.MULTIPOLYGON:
        func = _transpose_chunk_nest_3
    else:
        raise ValueError(f"Unexpected extension type name {extension_type_name}")

    return ChunkedArray([func(chunk) for chunk in column.chunks])


def _transpose_coords(arr: Array) -> Array:
    if DataType.is_fixed_size_list(arr.type):
        return arr

    if arr.type.num_fields == 2:
        x = struct_field(arr, [0]).to_numpy()
        y = struct_field(arr, [1]).to_numpy()
        coords = np.column_stack([x, y]).ravel("C")
        return fixed_size_list_array(coords, 2)

    if arr.type.num_fields == 3:
        x = struct_field(arr, [0]).to_numpy()
        y = struct_field(arr, [1]).to_numpy()
        z = struct_field(arr, [2]).to_numpy()
        coords = np.column_stack([x, y, z]).ravel("C")
        return fixed_size_list_array(coords, 3)

    raise ValueError(f"Expected struct with 2 or 3 fields, got {arr.type.num_fields}")


def _transpose_chunk_nest_0(arr: Array) -> Array:
    return _map_coords_nest_0(arr, _transpose_coords)


def _transpose_chunk_nest_1(arr: Array) -> Array:
    return _map_coords_nest_1(arr, _transpose_coords)


def _transpose_chunk_nest_2(arr: Array) -> Array:
    return _map_coords_nest_2(arr, _transpose_coords)


def _transpose_chunk_nest_3(arr: Array) -> Array:
    return _map_coords_nest_3(arr, _transpose_coords)



"""Handle GeoArrow tables with WKB-encoded geometry."""

import json

import numpy as np
from arro3.core import Field, Table
from geoarrow.rust.core import GeoChunkedArray, get_type_id

from lonboard._constants import EXTENSION_NAME, OGC_84
from lonboard._geoarrow.utils import is_primitive_geoarrow
from lonboard._utils import get_geometry_column_index

# Here we intentionally check geometries in a specific order.
# Starting from polygons, then linestrings, then points,
# so that the order of generated layers is polygon, then path then scatterplot.
# This ensures that points are rendered on top and polygons on the bottom.
#
# We can merge Point/LineString/Polygon and their Multi* counterparts, but we don't
# currently merge across dimension
TYPE_ID_ORDERING = [
    (16, 13),  # MultiPolygon Z, Polygon Z
    (6, 3),  # MultiPolygon, Polygon
    (15, 12),  # MultiLineString Z, LineString Z
    (5, 2),  # MultiLineString, LineString
    (14, 11),  # MultiPoint Z, Point Z
    (4, 1),  # MultiPoint, Point
]


def parse_serialized_table(table: Table) -> list[Table]:
    """Parse a table with a serialized WKB/WKT column into GeoArrow-native geometries.

    If no columns are WKB/WKT-encoded, returns the input. Note that WKB columns must be
    tagged with an extension name of `geoarrow.wkb` or `ogc.wkb`. WKT columns must be
    tagged with an extension name of `geoarrow.wkt`.

    Returns one table per GeoArrow geometry type. E.g. if the input has points, lines,
    and polygons, then it returns three tables. Point/MultiPoint,
    LineString/MultiLineString, Polygon/MultiPolygon are each combined into a single
    table type.
    """
    table = parse_geoparquet_table(table)
    field_idx = get_geometry_column_index(table.schema)

    # For non-geometry table, just return as-is
    if field_idx is None:
        return [table]

    field = table.field(field_idx)
    column = table.column(field_idx)

    # For native GeoArrow input, return table as-is
    # We need to downcast wkb/wkt/geometry
    if is_primitive_geoarrow(field):
        return [table]

    type_ids = get_type_id(column).read_all()
    unique_type_ids = {int(type_id) for type_id in np.asarray(type_ids)}
    if any(type_id >= 20 for type_id in unique_type_ids):
        raise ValueError(
            "Lonboard does not currently support M dimensional geometries.",
        )
    if any(type_id % 10 == 7 for type_id in unique_type_ids):
        raise ValueError(
            "Lonboard does not currently support GeometryCollections.",
        )

    # If there's a single type id, then we know we can downcast to that specific type.
    # In this case the input geometry is either WKB, WKT, or Geometry.
    if len(unique_type_ids) == 1:
        return [single_type_id_downcast(table)]

    type_ids_np = np.asarray(type_ids)

    parsed_tables: list[Table] = []
    concatted_batch = table.combine_chunks().to_batches()[0]

    for type_id_pair in TYPE_ID_ORDERING:
        if unique_type_ids.intersection(type_id_pair):
            indices = np.where(
                np.logical_or(
                    type_ids_np == type_id_pair[0],
                    type_ids_np == type_id_pair[1],
                ),
            )[0]
            assert len(indices) > 0, "Expected to find some geometries of this type."
            selected = concatted_batch.take(indices)
            parsed_tables.append(
                single_type_id_downcast(Table.from_batches([selected])),
            )

    return parsed_tables


def single_type_id_downcast(table: Table) -> Table:
    """Downcast table with single geometry type.

    Downcast a table whose geometry column is known to contain (serialized) geometries
    of only a single type... to a table whose geometry column has a primitive GeoArrow
    type.
    """
    field_idx = get_geometry_column_index(table.schema)
    assert field_idx is not None, "Expected a geometry column in the table."
    orig_field = table.schema.field(field_idx)

    chunked_geo_arr = GeoChunkedArray.from_arrow(table.column(field_idx))
    downcasted_array = chunked_geo_arr.downcast(coord_type="interleaved")
    downcasted_field = Field.from_arrow(downcasted_array.type)
    assert is_primitive_geoarrow(downcasted_field), (
        "Inside of single_type_id_downcast, expected to be able to downcast to a single primitive GeoArrow type",
    )

    return table.set_column(
        field_idx,
        downcasted_field.with_name(orig_field.name),
        downcasted_array,
    )


def parse_geoparquet_table(table: Table) -> Table:
    """Parse GeoParquet table metadata, assigning it to GeoArrow metadata."""
    # If a column already has geoarrow metadata, don't parse from GeoParquet metadata
    if get_geometry_column_index(table.schema) is not None:
        return table

    schema_metadata = table.schema.metadata or {}
    geo_metadata_bytes = schema_metadata.get(b"geo")
    if not geo_metadata_bytes:
        return table

    try:
        geo_metadata = json.loads(geo_metadata_bytes)
    except json.JSONDecodeError:
        return table

    primary_column = geo_metadata["primary_column"]
    column_meta = geo_metadata["columns"][primary_column]
    column_idxes = [
        idx for idx, name in enumerate(table.column_names) if name == primary_column
    ]
    assert len(column_idxes) == 1, f"Expected one column with name {primary_column}"
    column_idx = column_idxes[0]
    if column_meta["encoding"] == "WKB":
        existing_field = table.schema.field(column_idx)
        existing_column = table.column(column_idx)
        crs_metadata = {"crs": column_meta.get("crs", OGC_84.to_json_dict())}
        metadata: dict[bytes, bytes] = {
            b"ARROW:extension:name": EXTENSION_NAME.WKB,
            b"ARROW:extension:metadata": json.dumps(crs_metadata).encode(),
        }
        new_field = existing_field.with_metadata(metadata)
        table = table.set_column(column_idx, new_field, existing_column)

    return table



"""New traits for our custom serialization.

Refer to https://traitlets.readthedocs.io/en/stable/defining_traits.html for
documentation on how to define new traitlet types.
"""

# ruff: noqa: ARG002, C901, D102, D107, PLR0912, SLF001, UP031

from __future__ import annotations

import sys
import warnings
from typing import TYPE_CHECKING, Any, NoReturn, TypeVar
from typing import cast as type_cast
from urllib.parse import urlparse

import numpy as np
import traitlets
from arro3.core import (
    Array,
    ChunkedArray,
    DataType,
    Field,
    Table,
    fixed_size_list_array,
)
from traitlets import TraitError, Undefined
from traitlets.utils.descriptions import class_of, describe

from lonboard._constants import EXTENSION_NAME
from lonboard._environment import DEFAULT_HEIGHT
from lonboard._geoarrow.box_to_polygon import parse_box_encoded_table
from lonboard._serialization import (
    ACCESSOR_SERIALIZATION,
    TABLE_SERIALIZATION,
    serialize_view_state,
)
from lonboard._utils import get_geometry_column_index
from lonboard._vendor.matplotlib.colors import _to_rgba_no_colorcycle
from lonboard.models import ViewState

if TYPE_CHECKING:
    import pandas as pd
    from traitlets import HasTraits
    from traitlets.traitlets import TraitType
    from traitlets.utils.sentinel import Sentinel

    from lonboard._layer import BaseArrowLayer

DEFAULT_INITIAL_VIEW_STATE = {
    "latitude": 10,
    "longitude": 0,
    "zoom": 0.5,
    "bearing": 0,
    "pitch": 0,
}


class FixedErrorTraitType(traitlets.TraitType):
    """A custom subclass of traitlets.TraitType.

    This is because its `error` method ignores the `info` passed in. See
    https://github.com/developmentseed/lonboard/issues/71 and
    https://github.com/ipython/traitlets/pull/884.
    """

    def error(
        self,
        obj: HasTraits | None,
        value: Any,
        error: Exception | None = None,
        info: str | None = None,
    ) -> NoReturn:
        """Raise a TraitError.

        Parameters
        ----------
        obj : HasTraits or None
            The instance which owns the trait. If not
            object is given, then an object agnostic
            error will be raised.
        value : any
            The value that caused the error.
        error : Exception (default: None)
            An error that was raised by a child trait.
            The arguments of this exception should be
            of the form ``(value, info, *traits)``.
            Where the ``value`` and ``info`` are the
            problem value, and string describing the
            expected value. The ``traits`` are a series
            of :class:`TraitType` instances that are
            "children" of this one (the first being
            the deepest).
        info : str (default: None)
            A description of the expected value. By
            default this is infered from this trait's
            ``info`` method.

        """
        if error is not None:
            # handle nested error
            error.args += (self,)
            if self.name is not None:
                # this is the root trait that must format the final message
                chain = " of ".join(describe("a", t) for t in error.args[2:])
                if obj is not None:
                    error.args = (
                        "The '{}' trait of {} instance contains {} which "
                        "expected {}, not {}.".format(
                            self.name,
                            describe("an", obj),
                            chain,
                            error.args[1],
                            describe("the", error.args[0]),
                        ),
                    )
                else:
                    error.args = (
                        "The '{}' trait contains {} which expected {}, not {}.".format(
                            self.name,
                            chain,
                            error.args[1],
                            describe("the", error.args[0]),
                        ),
                    )
            raise error
        # this trait caused an error
        if self.name is None:
            # this is not the root trait
            raise TraitError(value, info or self.info(), self)
        # this is the root trait
        if obj is not None:
            e = "The '{}' trait of {} instance expected {}, not {}.".format(
                self.name,
                class_of(obj),
                # CHANGED:
                # Use info if provided
                info or self.info(),
                describe("the", value),
            )
        else:
            e = "The '{}' trait expected {}, not {}.".format(
                self.name,
                # CHANGED:
                # Use info if provided
                info or self.info(),
                describe("the", value),
            )
        raise TraitError(e)


class ArrowTableTrait(FixedErrorTraitType):
    """A trait to validate input for a geospatial Arrow-backed table.

    Allowed input includes:

    - A pyarrow [`Table`][pyarrow.Table] or containing a geometry column with [GeoArrow metadata](https://geoarrow.org/extension-types).
    - Any GeoArrow table from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = None
    info_text = (
        "a table-like Arrow object, such as a pyarrow or arro3 Table or "
        "RecordBatchReader"
    )

    def __init__(
        self: TraitType,
        *args: Any,
        allowed_geometry_types: set[EXTENSION_NAME] | None = None,
        allowed_dimensions: set[int] | None = None,
        geometry_required: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(
            sync=True,
            allowed_geometry_types=allowed_geometry_types,
            allowed_dimensions=allowed_dimensions,
            geometry_required=geometry_required,
            **TABLE_SERIALIZATION,
        )

    def validate(self, obj: BaseArrowLayer, value: Any) -> Table:
        if not isinstance(value, Table):
            self.error(obj, value)

        allowed_geometry_types = self.metadata.get("allowed_geometry_types")
        allowed_geometry_types = type_cast(
            "set[bytes] | None",
            allowed_geometry_types,
        )

        allowed_dimensions = self.metadata.get("allowed_dimensions")
        allowed_dimensions = type_cast("set[int] | None", allowed_dimensions)

        geom_col_idx = get_geometry_column_index(value.schema)

        geometry_required = self.metadata.get("geometry_required")
        if geometry_required and geom_col_idx is None:
            return self.error(obj, value, info="geometry column in table")

        # No restriction on the allowed geometry types in this table
        if allowed_geometry_types:
            # If we allow polygons as input, then we also allow geoarrow.box.
            # Convert boxes to Polygons
            if EXTENSION_NAME.POLYGON in allowed_geometry_types:
                value = parse_box_encoded_table(value)

            geometry_extension_type = value.schema.field(geom_col_idx).metadata.get(
                b"ARROW:extension:name",
            )

            if (
                allowed_geometry_types
                and geometry_extension_type not in allowed_geometry_types
            ):
                allowed_types_str = ", ".join(map(str, allowed_geometry_types))
                msg = (
                    f"Expected one of {allowed_types_str} geometry types, "
                    f"got {geometry_extension_type}."
                )
                self.error(obj, value, info=msg)

        if allowed_dimensions:
            typ = value.column(geom_col_idx).type
            while DataType.is_list(typ):
                value_type = typ.value_type
                assert value_type is not None
                typ = value_type

            assert DataType.is_fixed_size_list(typ)
            if typ.list_size not in allowed_dimensions:
                msg = " or ".join(map(str, list(allowed_dimensions)))
                self.error(obj, value, info=f"{msg}-dimensional points")

        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class ColorAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl color accessor.

    Various input is allowed:

    - A `list` or `tuple` with three or four integers, ranging between 0 and 255
      (inclusive). This will be used as the color for all objects.
    - A `str` representing a hex color or "well known" color interpretable by
      [matplotlib.colors.to_rgba][matplotlib.colors.to_rgba]. See also matplotlib's
      [list of named
      colors](https://matplotlib.org/stable/gallery/color/named_colors.html#base-colors).
    - A numpy `ndarray` with two dimensions and data type [`np.uint8`][numpy.uint8]. The
      size of the second dimension must be `3` or `4`, and will correspond to either RGB
      or RGBA colors.
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The inner
      size of the fixed size list must be `3` or `4` and its child must have type
      `uint8`.
    - Any Arrow fixed size list array from a library that implements the [Arrow
      PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).

    You can use helpers in the [`lonboard.colormap`][lonboard.colormap] module (i.e.
    [`apply_continuous_cmap`][lonboard.colormap.apply_continuous_cmap]) to simplify
    constructing numpy arrays for color values.
    """

    default_value = (0, 0, 0)
    info_text = (
        "a tuple or list representing an RGB(A) color or numpy ndarray or "
        "pyarrow FixedSizeList representing an array of RGB(A) colors"
    )

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if not np.issubdtype(value.dtype, np.uint8):
            self.error(obj, value, info="Color array must be uint8 type.")

        if value.ndim != 2:
            self.error(obj, value, info="Color array must have 2 dimensions.")

        list_size = value.shape[1]
        if list_size not in (3, 4):
            self.error(
                obj,
                value,
                info="Color array must have 3 or 4 as its second dimension.",
            )

        return ChunkedArray([fixed_size_list_array(value.ravel("C"), list_size)])

    def validate(self, obj: BaseArrowLayer, value: Any) -> tuple | list | ChunkedArray:
        if isinstance(value, (tuple, list)):
            if len(value) < 3 or len(value) > 4:
                self.error(obj, value, info="3 or 4 values if passed a tuple or list")

            if any(not isinstance(v, int) for v in value):
                self.error(
                    obj,
                    value,
                    info="all values to be integers if passed a tuple or list",
                )

            if any(v < 0 or v > 255 for v in value):
                self.error(
                    obj,
                    value,
                    info="values between 0 and 255",
                )

            return value

        if isinstance(value, str):
            try:
                c = _to_rgba_no_colorcycle(value)  # type: ignore
            except ValueError:
                return self.error(
                    obj,
                    value,
                    info=(
                        "Color string must be a named color or hex string interpretable"
                        " by matplotlib.colors.to_rgba."
                    ),
                )

            return tuple(map(int, (np.array(c) * 255).astype(np.uint8)))

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_fixed_size_list(value.type):
            self.error(obj, value, info="Color Arrow array must be a FixedSizeList.")

        if value.type.list_size not in (3, 4):
            self.error(
                obj,
                value,
                info=(
                    "Color Arrow array must have a FixedSizeList inner size of 3 or 4."
                ),
            )

        value_type = value.type.value_type
        assert value_type is not None
        if not DataType.is_uint8(value_type):
            self.error(obj, value, info="Color Arrow array must have a uint8 child.")

        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class FloatAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl float accessor.

    Various input is allowed:

    - An `int` or `float`. This will be used as the value for all objects.
    - A numpy `ndarray` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index.
    - A pandas `Series` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index.
    - A pyarrow [`FloatArray`][pyarrow.FloatArray], [`DoubleArray`][pyarrow.DoubleArray]
      or [`ChunkedArray`][pyarrow.ChunkedArray] containing either a `FloatArray` or
      `DoubleArray`. Each value in the array will be used as the value for the object at
      the same row index.
    - Any Arrow floating point array from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
      This includes data structures from
      [`geoarrow-rust`](https://geoarrow.github.io/geoarrow-rs/python/latest/).
    """

    default_value = float(0)
    info_text = (
        "a float value or numpy ndarray or Arrow array representing an array of floats"
    )

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _pandas_to_numpy(self, obj: BaseArrowLayer, value: pd.Series) -> np.ndarray:
        """Cast pandas Series to numpy ndarray."""
        return np.asarray(value)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if not np.issubdtype(value.dtype, np.number):
            self.error(obj, value, info="numeric dtype")

        # TODO: should we always be casting to float32? Should it be
        # possible/allowed to pass in ~int8 or a data type smaller than float32?
        return ChunkedArray([value.astype(np.float32)])

    def validate(self, obj: BaseArrowLayer, value: Any) -> float | ChunkedArray:
        if isinstance(value, (int, float)):
            return float(value)

        # pandas Series
        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            value = self._pandas_to_numpy(obj, value)

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_numeric(value.type):
            self.error(
                obj,
                value,
                info="Float Arrow array must be a numeric type.",
            )

        value = value.cast(DataType.float32())
        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class TextAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl text accessor.

    Various input is allowed:

    - A `str`. This will be used as the value for all objects.
    - A numpy `ndarray` with a string data type. Each value in the array will be used as
      the value for the object at the same row index.
    - A pandas `Series` with a string data type. Each value in the array will be used as
      the value for the object at the same row index.
    - A pyarrow [`StringArray`][pyarrow.StringArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing a `StringArray`. Each value in
      the array will be used as the value for the object at the same row index.
    """

    default_value = ""
    info_text = (
        "a string value or numpy ndarray or pandas Series or Arrow array representing"
        " an array of strings"
    )

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def pandas_to_arrow(self, obj: BaseArrowLayer, value: pd.Series) -> ChunkedArray:
        """Cast pandas Series to arrow array."""
        try:
            import pyarrow as pa
        except ImportError as e:
            raise ImportError(
                "pyarrow is a required dependency when passing in a pandas series",
            ) from e

        return ChunkedArray([pa.array(value)])

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        try:
            import pyarrow as pa
        except ImportError as e:
            raise ImportError(
                "pyarrow is a required dependency when passing in a numpy string array",
            ) from e

        return ChunkedArray([pa.StringArray.from_pandas(value)])

    def validate(self, obj: BaseArrowLayer, value: Any) -> float | str | ChunkedArray:
        if isinstance(value, str):
            return value

        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            value = self.pandas_to_arrow(obj, value)
        elif isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if DataType.is_large_string(value.type):
            value = value.cast(DataType.string())

        if not DataType.is_string(value.type):
            self.error(
                obj,
                value,
                info="String Arrow array must be a string type.",
            )

        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class PointAccessor(FixedErrorTraitType):
    """A representation of a deck.gl point accessor.

    Various input is allowed:

    - A numpy `ndarray` with two dimensions and data type [`np.uint8`][numpy.uint8]. The
      size of the second dimension must be `2` or `3`, and will correspond to either XY
      or XYZ positions.
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The inner
      size of the fixed size list must be `2` or `3` and its child must be of floating
      point type.
    """

    default_value = (0, 0, 0)
    info_text = "a numpy ndarray or arrow FixedSizeList representing a point array"

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if value.ndim != 2:
            self.error(obj, value, info="Point array to have 2 dimensions")

        list_size = value.shape[1]
        if list_size not in (2, 3):
            self.error(
                obj,
                value,
                info="Point array to have 2 or 3 as its second dimension",
            )

        assert np.issubdtype(value.dtype, np.float64)
        array = fixed_size_list_array(value.ravel("C"), list_size)
        return ChunkedArray([array])

    def validate(
        self,
        obj: BaseArrowLayer,
        value: Any,
    ) -> tuple[int, ...] | list[int] | ChunkedArray:
        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_fixed_size_list(value.type):
            self.error(obj, value, info="Point arrow array to be a FixedSizeList")

        if value.type.list_size not in (2, 3):
            self.error(
                obj,
                value,
                info=(
                    "point arrow array to be a FixedSizeList with list size of 2 or 3"
                ),
            )

        value_type = value.type.value_type
        assert value_type is not None
        if not DataType.is_float64(value_type):
            self.error(
                obj,
                value,
                info="Point arrow array to have a float64 point child",
            )

        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class FilterValueAccessor(FixedErrorTraitType):
    """Validate input for `get_filter_value`.

    A trait to validate input for the `get_filter_value` accessor added by the
    [`DataFilterExtension`][lonboard.layer_extension.DataFilterExtension], which can
    have between 1 and 4 float values per row.

    Various input is allowed:

    - An `int` or `float`. This will be used as the value for all objects. The
      `filter_size` of the
      [`DataFilterExtension`][lonboard.layer_extension.DataFilterExtension] instance
      must be 1.
    - A one-dimensional numpy `ndarray` with a numeric data type. This will be casted to
      an array of data type [`np.float32`][numpy.float32]. Each value in the array will
      be used as the value for the object at the same row index. The `filter_size` of
      the [`DataFilterExtension`][lonboard.layer_extension.DataFilterExtension] instance
      must be 1.
    - A two-dimensional numpy `ndarray` with a numeric data type. This will be casted to
      an array of data type [`np.float32`][numpy.float32]. Each value in the array will
      be used as the value for the object at the same row index. The `filter_size` of
      the [`DataFilterExtension`][lonboard.layer_extension.DataFilterExtension] instance
      must match the size of the second dimension of the array.
    - A pandas `Series` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index. The `filter_size` of the
      [`DataFilterExtension`][lonboard.layer_extension.DataFilterExtension] instance
      must be 1.
    - A pyarrow [`FloatArray`][pyarrow.FloatArray], [`DoubleArray`][pyarrow.DoubleArray]
      or [`ChunkedArray`][pyarrow.ChunkedArray] containing either a `FloatArray` or
      `DoubleArray`. Each value in the array will be used as the value for the object at
      the same row index. The `filter_size` of the
      [`DataFilterExtension`][lonboard.layer_extension.DataFilterExtension] instance
      must be 1.

      Alternatively, you can pass any corresponding Arrow data structure from a library
      that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The child
      array of the fixed size list must be of floating point type. The `filter_size` of
      the [`DataFilterExtension`][lonboard.layer_extension.DataFilterExtension] instance
      must match the list size.

      Alternatively, you can pass any corresponding Arrow data structure from a library
      that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = float(0)
    info_text = (
        "a float value or numpy ndarray or Arrow array representing an array of floats"
    )

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _pandas_to_numpy(
        self,
        obj: BaseArrowLayer,
        value: Any,
        filter_size: int,
    ) -> np.ndarray:
        # Assert that filter_size == 1 for a pandas series.
        # Pandas series can technically contain Python list objects inside them, but
        # for simplicity we disallow that.
        if filter_size != 1:
            self.error(obj, value, info="filter_size==1 with pandas Series")

        # Cast pandas Series to numpy ndarray
        return np.asarray(value)

    def _numpy_to_arrow(
        self,
        obj: BaseArrowLayer,
        value: Any,
        filter_size: int,
    ) -> ChunkedArray:
        if not np.issubdtype(value.dtype, np.number):
            self.error(obj, value, info="numeric dtype")

        # Cast to float32
        value = value.astype(np.float32)

        if len(value.shape) == 1:
            if filter_size != 1:
                self.error(obj, value, info="filter_size==1 with 1-D numpy array")

            return ChunkedArray([value])

        if len(value.shape) != 2:
            self.error(obj, value, info="1-D or 2-D numpy array")

        if value.shape[1] != filter_size:
            self.error(
                obj,
                value,
                info=(
                    f"filter_size ({filter_size}) to match 2nd dimension of numpy array"
                ),
            )

        array = fixed_size_list_array(value.ravel("C"), filter_size)
        return ChunkedArray([array])

    def validate(
        self,
        obj: BaseArrowLayer,
        value: Any,
    ) -> float | tuple | list | ChunkedArray:
        # Find the data filter extension in the attributes of the parent object so we
        # can validate against the filter size.
        data_filter_extension = [
            ext
            for ext in obj.extensions
            if ext._extension_type == "data-filter"  # type: ignore
        ]
        assert len(data_filter_extension) == 1
        filter_size = data_filter_extension[0].filter_size  # type: ignore

        if isinstance(value, (int, float)):
            if filter_size != 1:
                self.error(obj, value, info="filter_size==1 with scalar value")

            return float(value)

        if isinstance(value, (tuple, list)):
            if filter_size != len(value):
                self.error(
                    obj,
                    value,
                    info=f"filter_size ({filter_size}) to match length of tuple/list",
                )

            if any(not isinstance(v, (int, float)) for v in value):
                self.error(
                    obj,
                    value,
                    info="all values in tuple or list to be numeric",
                )

            return value

        # pandas Series
        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            value = self._pandas_to_numpy(obj, value, filter_size)

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value, filter_size)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        # Allowed inputs are either a FixedSizeListArray or numeric array.
        # If not a fixed size list array, check for floating and cast to float32
        if not DataType.is_fixed_size_list(value.type):
            if filter_size != 1:
                self.error(
                    obj,
                    value,
                    info="filter_size==1 with non-FixedSizeList type arrow array",
                )

            if not DataType.is_floating(value.type):
                self.error(
                    obj,
                    value,
                    info="arrow array to be a floating point type",
                )

            return value.cast(DataType.float32())

        # We have a FixedSizeListArray
        if filter_size != value.type.list_size:
            self.error(
                obj,
                value,
                info=(
                    f"filter_size ({filter_size}) to match list size of "
                    "FixedSizeList arrow array"
                ),
            )

        value_type = value.type.value_type
        assert value_type is not None
        if not DataType.is_floating(value_type):
            self.error(
                obj,
                value,
                info="arrow array to have floating point child type",
            )

        # Cast values to float32
        value = value.cast(
            DataType.list(Field("", DataType.float32()), value.type.list_size),
        )
        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class NormalAccessor(FixedErrorTraitType):
    """A representation of a deck.gl "normal" accessor.

    This is primarily used in the [lonboard.PointCloudLayer].

    Acceptable inputs:
    - A `list` or `tuple` with three `int` or `float` values. This will be used as the
      normal for all objects.
    - A numpy ndarray with two dimensions and floating point type. The size of the
      second dimension must be 3, i.e. its shape must be `(N, 3)`.
    - a pyarrow `FixedSizeListArray` or `ChunkedArray` containing `FixedSizeListArray`s
      where the size of the inner fixed size list 3. The child array must have type
      float32.
    - Any Arrow array that matches the above restrictions from a library that implements
      the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = (0, 0, 1)
    info_text = (
        "List representing normal of all objects in [nx, ny, nz] or numpy ndarray or "
        "pyarrow FixedSizeList representing the normal of each object, in [nx, ny, nz]"
    )

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if not np.issubdtype(value.dtype, np.number):
            self.error(obj, value, info="normal array to have numeric type")

        if value.ndim != 2 or value.shape[1] != 3:
            self.error(obj, value, info="normal array to be 2D with shape (N, 3)")

        if not np.issubdtype(value.dtype, np.float32):
            warnings.warn(
                """Warning: Numpy array should be float32 type.
                Converting to float32 point Arrow array""",
            )
            value = value.astype(np.float32)

        array = fixed_size_list_array(value.ravel("C"), 3)
        return ChunkedArray([array])

    def validate(
        self,
        obj: BaseArrowLayer,
        value: Any,
    ) -> tuple[int, ...] | list[int] | ChunkedArray:
        if isinstance(value, (tuple, list)):
            if len(value) != 3:
                self.error(
                    obj,
                    value,
                    info="normal scalar to have length 3, (nx, ny, nz)",
                )

            if not all(isinstance(item, (int, float)) for item in value):
                self.error(
                    obj,
                    value,
                    info="all elements of normal scalar to be int or float type",
                )

            return value

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_fixed_size_list(value.type):
            self.error(obj, value, info="normal Arrow array to be a FixedSizeList.")

        if value.type.list_size != 3:
            self.error(
                obj,
                value,
                info=("normal Arrow array to have an inner size of 3."),
            )

        value_type = value.type.value_type
        assert value_type is not None
        if not DataType.is_floating(value_type):
            self.error(
                obj,
                value,
                info="Arrow array to be floating point type",
            )

        value = value.cast(DataType.list(Field("", DataType.float32()), 3))
        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class ViewStateTrait(FixedErrorTraitType):
    """Trait to validate view state input."""

    allow_none = True
    default_value = DEFAULT_INITIAL_VIEW_STATE

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.tag(sync=True, to_json=serialize_view_state)

    def validate(self, obj: Any, value: Any) -> None | ViewState:
        if value is None:
            return None

        if isinstance(value, ViewState):
            return value

        if isinstance(value, dict):
            value = {**DEFAULT_INITIAL_VIEW_STATE, **value}
            return ViewState(**value)

        self.error(obj, value)
        assert False


class DashArrayAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl dash accessor.

    Primarily used in
    [`PathStyleExtension`][lonboard.layer_extension.PathStyleExtension].

    Various input is allowed:

    - A `list` or `tuple` with 2 integers. This defines the dash size and gap size
      respectively.
    - A numpy `ndarray` with two dimensions and numeric data type. The size of the
      second dimension must be `2`.
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The inner
      size of the fixed size list must be `2`.
    - Any Arrow fixed size list array from a library that implements the [Arrow
      PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = (0, 0)
    info_text = (
        "a tuple or list or numpy ndarray or "
        "pyarrow FixedSizeList representing dash size and gap size."
    )

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if not np.issubdtype(value.dtype, np.number):
            self.error(obj, value, info="NumPy array must be uint8 type.")

        if value.ndim != 2:
            self.error(obj, value, info="NumPy array must have 2 dimensions.")

        list_size = value.shape[1]
        if list_size != 2:
            self.error(
                obj,
                value,
                info="NumPy array must have 2 as its second dimension.",
            )

        # Cast float64 to float32; leave other data types the same
        if np.issubdtype(value.dtype, np.float64):
            value = value.astype(np.float32)

        array = fixed_size_list_array(value.ravel("C"), list_size)
        return ChunkedArray([array])

    def validate(
        self,
        obj: BaseArrowLayer,
        value: Any,
    ) -> tuple[int, ...] | list[int] | ChunkedArray:
        if isinstance(value, (tuple, list)):
            if len(value) != 2:
                self.error(obj, value, info="2 value list only")

            if any(not isinstance(v, (int, float)) for v in value):
                self.error(
                    obj,
                    value,
                    info="all values to be int or float type if passed a tuple or list",
                )

            return value

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_fixed_size_list(value.type):
            self.error(obj, value, info="Arrow array must be a FixedSizeList.")

        if value.type.list_size != 2:
            self.error(
                obj,
                value,
                info="Arrow array must have a FixedSizeList inner size of 2.",
            )

        value_type = value.type.value_type
        assert value_type is not None
        if not (
            DataType.is_integer(value_type)
            or DataType.is_signed_integer(value_type)
            or DataType.is_floating(value_type)
        ):
            self.error(obj, value, info="Arrow array to have a numeric type child.")

        # Cast float64 to float32; leave other data types the same
        if DataType.is_float64(value_type):
            value = value.cast(DataType.list(DataType.float32(), value.type.list_size))

        return value.rechunk(max_chunksize=obj._rows_per_chunk)


class BasemapUrl(traitlets.Unicode):
    """Validation for basemap url."""

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)

    def validate(self, obj: HasTraits | None, value: Any) -> Any:
        value = super().validate(obj, value)

        try:
            parsed = urlparse(value)
        except:  # noqa: E722
            self.error(obj, value, info="to be a URL")

        if not parsed.scheme.startswith("http"):
            self.error(obj, value, info="to be a HTTP(s) URL")

        return value


T = TypeVar("T")


# TODO: switch to
# class VariableLengthTuple(traitlets.Container[tuple[T, ...]])
# When we can upgrade to traitlets 5.10 (depends on Colab upgrading)
class VariableLengthTuple(traitlets.Container):
    """An instance of a Python tuple with variable numbers of elements of the same type."""

    klass = list  # type:ignore[assignment]
    _cast_types: Any = (tuple,)

    def __init__(
        self,
        trait: T | Sentinel = None,
        default_value: tuple[T] | Sentinel | None = Undefined,
        minlen: int = 0,
        maxlen: int = sys.maxsize,
        **kwargs: Any,
    ) -> None:
        """Create a tuple trait type.

        The default value is created by doing ``list(default_value)``,
        which creates a copy of the ``default_value``.

        ``trait`` can be specified, which restricts the type of elements
        in the container to that TraitType.

        If only one arg is given and it is not a Trait, it is taken as
        ``default_value``:

        ``c = List([1, 2, 3])``

        Parameters
        ----------
        trait : TraitType [ optional ]
            the type for restricting the contents of the Container.
            If unspecified, types are not checked.
        default_value : SequenceType [ optional ]
            The default value for the Trait.  Must be list/tuple/set, and
            will be cast to the container type.
        minlen : Int [ default 0 ]
            The minimum length of the input list
        maxlen : Int [ default sys.maxsize ]
            The maximum length of the input list
        kwargs:
            passed on to traitlets.Container.

        """
        self._maxlen = maxlen
        self._minlen = minlen
        super().__init__(trait=trait, default_value=default_value, **kwargs)

    def length_error(self, obj: Any, value: Any) -> None:
        e = "The '%s' trait of %s instance must be of length %i <= L <= %i" % (
            self.name,
            class_of(obj),
            self._minlen,
            self._maxlen,
        )
        e += f", but a value of {value} was specified."
        raise TraitError(e)

    def validate_elements(self, obj: Any, value: Any) -> Any:
        length = len(value)
        if length < self._minlen or length > self._maxlen:
            self.length_error(obj, value)

        trait = self._trait

        validated = []
        for v in value:
            try:
                v = trait._validate(obj, v)  # noqa: PLW2901
            except TraitError as error:  # noqa: PERF203
                self.error(obj, v, error)
            else:
                validated.append(v)

        return tuple(validated)


class HeightTrait(FixedErrorTraitType):
    """Trait to validate map height input."""

    allow_none = True
    default_value = DEFAULT_HEIGHT

    def __init__(
        self: TraitType,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.tag(sync=True)

    def validate(self, obj: Any, value: Any) -> str:
        if isinstance(value, int):
            return f"{value}px"

        if isinstance(value, str):
            return value

        self.error(obj, value)
        assert False



from __future__ import annotations

from io import StringIO
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, TextIO, overload

import ipywidgets
import traitlets
import traitlets as t
from ipywidgets import CallbackDispatcher
from ipywidgets.embed import dependency_state, embed_minimal_html

from lonboard._base import BaseAnyWidget
from lonboard._layer import BaseLayer
from lonboard._viewport import compute_view
from lonboard.basemap import CartoBasemap
from lonboard.traits import (
    DEFAULT_INITIAL_VIEW_STATE,
    BasemapUrl,
    HeightTrait,
    VariableLengthTuple,
    ViewStateTrait,
)

if TYPE_CHECKING:
    import sys
    from collections.abc import Callable, Sequence

    from IPython.display import HTML  # type: ignore

    from lonboard.types.map import MapKwargs

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


# bundler yields lonboard/static/{index.js,styles.css}
bundler_output_dir = Path(__file__).parent / "static"

# HTML template to override exported map as 100% height
_HTML_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>{title}</title>
</head>
<style>
    html {{ height: 100%; }}
    body {{ height: 100%; overflow: hidden;}}
    .widget-subarea {{ height: 100%; }}
    .jupyter-widgets-disconnected {{ height: 100%; }}
</style>
<body>
{snippet}
</body>
</html>
"""


class Map(BaseAnyWidget):
    """The top-level class used to display a map in a Jupyter Widget.

    **Example:**

    ```py
    import geopandas as gpd
    from lonboard import Map, ScatterplotLayer, SolidPolygonLayer

    # A GeoDataFrame with Point geometries
    point_gdf = gpd.GeoDataFrame()
    point_layer = ScatterplotLayer.from_geopandas(
        point_gdf,
        get_fill_color=[255, 0, 0],
    )

    # A GeoDataFrame with Polygon geometries
    polygon_gdf = gpd.GeoDataFrame()
    polygon_layer = SolidPolygonLayer.from_geopandas(
        gdf,
        get_fill_color=[255, 0, 0],
    )

    m = Map([point_layer, polygon_layer])
    ```
    """

    def __init__(
        self,
        layers: BaseLayer | Sequence[BaseLayer],
        **kwargs: Unpack[MapKwargs],
    ) -> None:
        """Create a new Map.

        Aside from the `layers` argument, pass keyword arguments for any other attribute
        defined in this class.

        Args:
            layers: One or more layers to render on this map.

        Keyword Args:
            kwargs: Passed on to class variables.

        Returns:
            A Map object.

        """
        if isinstance(layers, BaseLayer):
            layers = [layers]

        def _handle_anywidget_dispatch(
            widget: ipywidgets.Widget,  # noqa: ARG001
            msg: dict,
            buffers: list[bytes],  # noqa: ARG001
        ) -> None:
            if msg.get("kind") == "on-click":
                self._click_handlers(tuple(msg.get("coordinate")))

        super().__init__(layers=layers, **kwargs)
        self._click_handlers = CallbackDispatcher()
        self.on_msg(_handle_anywidget_dispatch)
        self.layout.height = "100%"
        self.layout.width = "100%"

    def on_click(self, callback: Callable, *, remove: bool = False) -> None:
        """Register a callback to execute when the map is clicked.

        The callback will be called with one argument, a tuple of the coordinate
        clicked (x,y)/(Longitude/Latitude).


        Args:
            callback: function callback to pass to click handler.
            remove: bool (optional)
                Set to true to remove the callback from the list of callbacks.

        !!! note
            If the map is zoomed to a very large scale and can see the earth wrapped
            around, it is possible the coordinate's x/Longitude value may be greater
            than or less than expected.  Example: If you can see Paris, France three
            times in the map, and you click on the Paris in the middle, it will show an
            X coordinate of 2, but the Paris on the left of the map will report an X
            coordinate of -358, and the Paris on the right of the map will report an
            X coordinate of 362.

        """
        self._click_handlers.register_callback(callback, remove=remove)
        self._has_click_handlers = len(self._click_handlers.callbacks) > 0

    _esm = bundler_output_dir / "index.js"
    _css = bundler_output_dir / "index.css"

    view_state = ViewStateTrait()
    """
    The view state of the map.

    - Type: [`ViewState`][lonboard.models.ViewState]
    - Default: Automatically inferred from the data passed to the map.

    You can initialize the map to a specific view state using this property:

    ```py
    Map(
        layers,
        view_state={"longitude": -74.0060, "latitude": 40.7128, "zoom": 7}
    )
    ```

    !!! note

        The properties of the view state are immutable. Use
        [`set_view_state`][lonboard.Map.set_view_state] to modify a map's view state
        once it's been initially rendered.

    """
    _has_click_handlers = t.Bool(default_value=False, allow_none=False).tag(sync=True)
    """
    Indicates if a click handler has been registered.
    """

    height = HeightTrait().tag(sync=True)
    """Height of the map in pixels, or valid CSS height property.

    This API is not yet stabilized and may change in the future.
    """

    layers = VariableLengthTuple(t.Instance(BaseLayer)).tag(
        sync=True,
        **ipywidgets.widget_serialization,
    )
    """One or more `Layer` objects to display on this map.
    """

    show_tooltip = t.Bool(default_value=False).tag(sync=True)
    """
    Whether to render a tooltip on hover on the map.

    - Type: `bool`
    - Default: `False`
    """

    show_side_panel = t.Bool(default_value=True).tag(sync=True)
    """
    Whether to render a side panel upon clicking on elements.

    - Type: `bool`
    - Default: `True`
    """

    picking_radius = t.Int(5).tag(sync=True)
    """
    Extra pixels around the pointer to include while picking (such as for a tooltip).

    This is helpful when rendered objects are difficult to target, for example
    irregularly shaped icons, small moving circles or interaction by touch.

    - Type: `int`
    - Default: `5`
    """

    basemap_style = BasemapUrl(CartoBasemap.PositronNoLabels)
    """
    A URL to a MapLibre-compatible basemap style.

    Various styles are provided in [`lonboard.basemap`](https://developmentseed.org/lonboard/latest/api/basemap/).

    - Type: `str`, holding a URL hosting a basemap style.
    - Default
      [`lonboard.basemap.CartoBasemap.PositronNoLabels`][lonboard.basemap.CartoBasemap.PositronNoLabels]
    """

    custom_attribution = t.Union(
        [
            t.Unicode(allow_none=True),
            VariableLengthTuple(t.Unicode(allow_none=False)),
        ],
    ).tag(sync=True)
    """
    Custom attribution to display on the map.

    This attribute supports the same format as the `attribution` property in the
    Maplibre API.

    - Type: `str` or `List[str]`
    - Default: `None`

    You can provide either a single string or a list of strings for custom attributions.
    If an attribution value is set in the map style, it will be displayed in addition to
    this custom attribution.

    **Example:**

        ```py
        m = Map(
            layers,
            custom_attribution="Development Seed"
        )
        ```

    **Example:**

        ```py
        m = Map(
            layers,
            custom_attribution=["Development Seed", "OpenStreetMap"]
        )
        ```
    """

    # TODO: We'd prefer a "Strict union of bool and float" but that doesn't
    # work here because `Union[bool, float]` would coerce `1` to `True`, which we don't
    # want, and `Union[float, bool]` would coerce `True` to `1`, which we also don't
    # want.
    # In the future we could create a custom trait for this if asked for.
    use_device_pixels = t.Any(allow_none=True, default_value=None).tag(sync=True)
    """Controls the resolution of the drawing buffer used for rendering.

    Setting this to `false` or a number <= 1 will improve performance on high resolution
    displays.

    **Note**: This parameter must be passed to the `Map()` constructor. It cannot be
    changed once the map has been created.

    The available options are:

    - `true`: Device (physical) pixels resolution is used for rendering, this resolution
      is defined by `window.devicePixelRatio`. On Retina/HD systems this resolution is
      usually twice as big as CSS pixels resolution.
    - `false`: CSS pixels resolution (equal to the canvas size) is used for rendering.
    - `Number` (Experimental): Specified Number is used as a custom ratio (drawing
      buffer resolution to CSS pixel resolution) to determine drawing buffer size, a
      value less than one uses resolution smaller than CSS pixels, gives better
      performance but produces blurry images, a value greater than one uses resolution
      bigger than CSS pixels resolution (canvas size), produces sharp images but at a
      lower performance.

    - Type: `float`, `int` or `bool`
    - Default: `true`
    """

    parameters = t.Any(allow_none=True, default_value=None).tag(sync=True)
    """GPU parameters to pass to deck.gl.

    **This is an advanced API. The vast majority of users should not need to touch this
    setting.**

    !!! Note

        The docstring below is copied from upstream deck.gl documentation. Any usage of
        `GL` refers to the constants defined in [`@luma.gl/constants`
        here](https://github.com/visgl/luma.gl/blob/master/modules/constants/src/webgl-constants.ts),
        which comes from the [MDN docs
        here](https://developer.mozilla.org/en-US/docs/Web/API/WebGL_API/Constants).

        In place of any `GL` constant, you can use the underlying integer it represents.
        For example, instead of the JS

        ```
        depthFunc: GL.LEQUAL
        ```

        referring to the [MDN
        docs](https://developer.mozilla.org/en-US/docs/Web/API/WebGL_API/Constants#depth_or_stencil_tests),
        you should use

        ```
        depthFunc: 0x0203
        ```

        Note that these parameters do not yet work with integer keys. If you would like
        to use integer keys, open an issue.

    Expects an object with GPU parameters. Before each frame is rendered, this object
    will be passed to luma.gl's `setParameters` function to reset the GPU context
    parameters, e.g. to disable depth testing, change blending modes etc. The default
    parameters set by `Deck` on initialization are the following:

    ```js
    {
      blend: true,
      blendFunc: [GL.SRC_ALPHA, GL.ONE_MINUS_SRC_ALPHA, GL.ONE, GL.ONE_MINUS_SRC_ALPHA],
      polygonOffsetFill: true,
      depthTest: true,
      depthFunc: GL.LEQUAL
    }
    ```

    Refer to the luma.gl
    [setParameters](https://github.com/visgl/luma.gl/blob/8.5-release/modules/gltools/docs/api-reference/parameter-setting.md)
    API for documentation on supported parameters and values.

    ```js
    import GL from '@luma.gl/constants';
    new Deck({
      // ...
      parameters: {
        blendFunc: [GL.ONE, GL.ONE, GL.ONE, GL.ONE],
        depthTest: false
      }
    });
    ```

    Notes:

    - Any GPU `parameters` prop supplied to individual layers will still override the
      global `parameters` when that layer is rendered.
    """

    selected_bounds = t.Tuple(
        t.Float(),
        t.Float(),
        t.Float(),
        t.Float(),
        allow_none=True,
        default_value=None,
    ).tag(sync=True)
    """
    Bounds selected by the user, represented as a tuple of floats ordered as

    ```
    (minx, miny, maxx, maxy)
    ```
    """

    def add_layer(
        self,
        layers: BaseLayer | Sequence[BaseLayer] | Map,
        *,
        focus: bool = False,
        reset_zoom: bool = False,
    ) -> None:
        """Add one or more new layers to the map.

        Examples:
        ```py
        from lonboard import viz

        m = viz(some_data)
        m.add_layer(viz(more_data), focus=True)
        ```

        Args:
            layers: New layers to add to the map. This can be:
                - a layer instance
                - a list or tuple of layer instances
                - another `Map` instance, in which case its layers will be added to this
                  map. This lets you pass the result of `viz` into this method.

            focus: If True, set the view state of the map based on the _newly-added_
                layers. Defaults to False.
            reset_zoom: If True, set the view state of the map based on _all_ layers.
                Defaults to False.

        Raises:
            ValueError: _description_

        """
        if focus and reset_zoom:
            raise ValueError("focus and reset_zoom may not both be set.")

        if isinstance(layers, Map):
            new_layers = layers.layers
        elif isinstance(layers, BaseLayer):
            new_layers = (layers,)
        else:
            new_layers = tuple(layers)

        self.layers += new_layers

        if focus:
            self.view_state = compute_view(new_layers)  # type: ignore

        elif reset_zoom:
            self.view_state = compute_view(self.layers)  # type: ignore

    def set_view_state(
        self,
        *,
        longitude: float | None = None,
        latitude: float | None = None,
        zoom: float | None = None,
        pitch: float | None = None,
        bearing: float | None = None,
    ) -> None:
        """Set the view state of the map.

        Any parameters that are unset will not be changed.

        Keyword Args:
            longitude: the new longitude to set on the map. Defaults to None.
            latitude: the new latitude to set on the map. Defaults to None.
            zoom: the new zoom to set on the map. Defaults to None.
            pitch: the new pitch to set on the map. Defaults to None.
            bearing: the new bearing to set on the map. Defaults to None.

        """
        view_state = (
            self.view_state._asdict()  # type: ignore
            if self.view_state is not None
            else DEFAULT_INITIAL_VIEW_STATE
        )

        if longitude is not None:
            view_state["longitude"] = longitude
        if latitude is not None:
            view_state["latitude"] = latitude
        if zoom is not None:
            view_state["zoom"] = zoom
        if pitch is not None:
            view_state["pitch"] = pitch
        if bearing is not None:
            view_state["bearing"] = bearing

        self.view_state = view_state

    def fly_to(  # noqa: PLR0913
        self,
        *,
        longitude: float,
        latitude: float,
        zoom: float,
        duration: int = 4000,
        pitch: float = 0,
        bearing: float = 0,
        curve: float | None = None,
        speed: float | None = None,
        screen_speed: float | None = None,
    ) -> None:
        r""" "Fly" the map to a new location.

        Args:
            longitude: The longitude of the new viewport.
            latitude: The latitude of the new viewport.
            zoom: The zoom of the new viewport.
            pitch: The pitch of the new viewport. Defaults to 0.
            bearing: The bearing of the new viewport. Defaults to 0.
            duration: The number of milliseconds for the viewport transition to take.
                Defaults to 4000.
            curve: The zooming "curve" that will occur along the flight path. Default
                `1.414`.
            speed: The average speed of the animation defined in relation to
                `curve`, it linearly affects the duration, higher speed returns smaller
                durations and vice versa. Default `1.2`.
            screen_speed: The average speed of the animation measured in screenfuls per
                second. Similar to speed it linearly affects the duration, when
                specified speed is ignored.

        """  # noqa: D210
        if not isinstance(longitude, (int, float)):
            raise TypeError(
                f"Expected longitude to be an int or float, got {type(longitude)}",
            )

        if not isinstance(latitude, (int, float)):
            raise TypeError(
                f"Expected latitude to be an int or float, got {type(latitude)}",
            )

        if not isinstance(zoom, (int, float)):
            raise TypeError(f"Expected zoom to be an int or float, got {type(zoom)}")

        data = {
            "type": "fly-to",
            "longitude": longitude,
            "latitude": latitude,
            "zoom": zoom,
            "pitch": pitch,
            "bearing": bearing,
            "transitionDuration": duration,
            "curve": curve,
            "speed": speed,
            "screenSpeed": screen_speed,
        }
        self.send(data)

    @overload
    def to_html(
        self,
        filename: None = None,
        title: str | None = None,
    ) -> str: ...

    @overload
    def to_html(
        self,
        filename: str | Path | TextIO | IO[str],
        title: str | None = None,
    ) -> None: ...

    def to_html(
        self,
        filename: str | Path | TextIO | IO[str] | None = None,
        title: str | None = None,
    ) -> str | None:
        """Save the current map as a standalone HTML file.

        Args:
            filename: where to save the generated HTML file.

        Keyword Args:
            title: A title for the exported map. This will show as the browser tab name.

        Returns:
            If `filename` is not passed, returns the HTML content as a `str`.

        """

        def inner(fp: str | Path | TextIO | IO[str]) -> None:
            original_height = self.height
            try:
                with self.hold_trait_notifications():
                    self.height = "100%"
                    embed_minimal_html(
                        fp,
                        views=[self],
                        title=title or "Lonboard export",
                        template=_HTML_TEMPLATE,
                        drop_defaults=False,
                        # Necessary to pass the state of _this_ specific map. Otherwise, the
                        # state of all known widgets will be included, ballooning the file size.
                        state=dependency_state((self), drop_defaults=False),
                    )
            finally:
                # If the map had a height before the HTML was generated, reset it.
                self.height = original_height

        if filename is None:
            with StringIO() as sio:
                inner(sio)
                return sio.getvalue()

        else:
            inner(filename)
            return None

    def as_html(self) -> HTML:
        """Render the current map as a static HTML file in IPython.

        !!! warning

            The primary, recommended way to display a map is by
            leaving it as the last line in a cell.

            ```py
            from lonboard import Map

            m = Map(layers=[])
            m
            ```

            This method exists to support environments that are unable to display
            Jupyter Widgets. Some aspects of Lonboard are unavailable with this display
            method. In particular, the map is unable to send any information back to
            Python. So [`selected_index`][lonboard.BaseArrowLayer.selected_index] will
            never be populated, for example.

        Returns:
            IPython HTML object.

        """
        from IPython.display import HTML  # type: ignore

        return HTML(self.to_html())

    @traitlets.default("view_state")
    def _default_initial_view_state(self) -> dict[str, Any]:
        return compute_view(self.layers)  # type: ignore



# Notes:
#
# - See module docstring of lonboard._layer for note on passing None as default value.

# ruff: noqa

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import ipywidgets
import traitlets
import traitlets as t
from arro3.core import DataType, Scalar

from lonboard._constants import EXTENSION_NAME, MIN_INTEGER_FLOAT32
from lonboard._layer import BaseArrowLayer
from lonboard._utils import timestamp_max_physical_value, timestamp_start_offset
from lonboard.experimental.traits import TimestampAccessor
from lonboard.traits import (
    ArrowTableTrait,
    ColorAccessor,
    FloatAccessor,
    PointAccessor,
    TextAccessor,
)

if TYPE_CHECKING:
    import sys

    import duckdb
    import geopandas as gpd
    import pyproj
    from arro3.core.types import ArrowStreamExportable
    from movingpandas import TrajectoryCollection

    from lonboard.types.layer import TripsLayerKwargs

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


class ArcLayer(BaseArrowLayer):
    """Render raised arcs joining pairs of source and target coordinates."""

    _layer_type = t.Unicode("arc").tag(sync=True)

    table = ArrowTableTrait(geometry_required=False)
    """A GeoArrow table.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.ScatterplotLayer.from_geopandas] instead.
    """

    great_circle = t.Bool(None, allow_none=True).tag(sync=True)
    """If `True`, create the arc along the shortest path on the earth surface.

    - Type: `bool`, optional
    - Default: `False`
    """

    num_segments = t.Int(None, allow_none=True).tag(sync=True)
    """The number of segments used to draw each arc.

    - Type: `int`, optional
    - Default: `50`
    """

    width_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'pixels'`
    """

    width_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The scaling multiplier for the width of each line.

    - Type: `float`, optional
    - Default: `1`
    """

    width_min_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """The minimum line width in pixels.

    - Type: `float`, optional
    - Default: `0`
    """

    width_max_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """The maximum line width in pixels.

    - Type: `float`, optional
    - Default: `None`
    """

    get_source_position = PointAccessor(None, allow_none=True)
    """Source position of each object
    """

    get_target_position = PointAccessor(None, allow_none=True)
    """Target position of each object
    """

    get_source_color = ColorAccessor(None, allow_none=True)
    """Source color of each object
    """

    get_target_color = ColorAccessor(None, allow_none=True)
    """Target color of each object
    """

    get_width = FloatAccessor(None, allow_none=True)
    """The line width of each object, in units specified by `widthUnits`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all paths.
        - If an array is provided, each value in the array will be used as the width for
          the path at the same row index.
    - Default: `1`.
    """

    get_height = FloatAccessor(None, allow_none=True)
    """Height color of each object
    """

    get_tilt = FloatAccessor(None, allow_none=True)
    """
    Use to tilt the arc to the side if you have multiple arcs with the same source and
    target positions.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all paths.
        - If an array is provided, each value in the array will be used as the width for
          the path at the same row index.
    - Default: `0`.
    """


class TextLayer(BaseArrowLayer):
    """Render text labels at given coordinates."""

    _layer_type = t.Unicode("text").tag(sync=True)

    table = ArrowTableTrait(allowed_geometry_types={EXTENSION_NAME.POINT})
    """A GeoArrow table with a Point or MultiPoint column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.ScatterplotLayer.from_geopandas] instead.
    """

    billboard = t.Bool(None, allow_none=True).tag(sync=True)
    """If `true`, the text always faces camera. Otherwise the text faces up (z).

    - Type: `bool`
    - Default: `True`
    """

    size_scale = t.Any(None, allow_none=True).tag(sync=True)
    """Text size multiplier.

    - Type: `float`.
    - Default: `1`
    """

    size_units = t.Any(None, allow_none=True).tag(sync=True)
    """The units of the size, one of `'meters'`, `'common'`, and `'pixels'`.
    default 'pixels'. See [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'pixels'`
    """

    size_min_pixels = t.Any(None, allow_none=True).tag(sync=True)
    """
    The minimum size in pixels. When using non-pixel `sizeUnits`, this prop can be used
    to prevent the icon from getting too small when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    size_max_pixels = t.Any(None, allow_none=True).tag(sync=True)
    """
    The maximum size in pixels. When using non-pixel `sizeUnits`, this prop can be used
    to prevent the icon from getting too big when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    # background = t.Bool(None, allow_none=True).tag(sync=True)
    # """Whether to render background for the text blocks.

    # - Type: `bool`
    # - Default: `False`
    # """

    get_background_color = ColorAccessor(None, allow_none=True)
    """Background color accessor.

    default [255, 255, 255, 255]
    """

    get_border_color = ColorAccessor(None, allow_none=True)
    """Border color accessor.

    default [0, 0, 0, 255]
    """

    get_border_width = FloatAccessor(None, allow_none=True)
    """Border width accessor.

    default 0
    """

    background_padding = t.Any(None, allow_none=True).tag(sync=True)
    """The padding of the background.

    - If an array of 2 is supplied, it is interpreted as `[padding_x, padding_y]` in
      pixels.
    - If an array of 4 is supplied, it is interpreted as `[padding_left, padding_top,
      padding_right, padding_bottom]` in pixels.

    default [0, 0, 0, 0]
    """

    character_set = t.Any(None, allow_none=True).tag(sync=True)
    """
    Specifies a list of characters to include in the font. If set to 'auto', will be
    automatically generated from the data set.

    default (ASCII characters 32-128)
    """

    font_family = t.Any(None, allow_none=True).tag(sync=True)
    """CSS font family

    default 'Monaco, monospace'
    """

    font_weight = t.Any(None, allow_none=True).tag(sync=True)
    """CSS font weight

    default 'normal'
    """

    line_height = t.Any(None, allow_none=True).tag(sync=True)
    """
    A unitless number that will be multiplied with the current text size to set the line
    height.
    """

    outline_width = t.Any(None, allow_none=True).tag(sync=True)
    """
    Width of outline around the text, relative to the text size. Only effective if
    `fontSettings.sdf` is `true`.

    default 0
    """

    outline_color = t.Any(None, allow_none=True).tag(sync=True)
    """
    Color of outline around the text, in `[r, g, b, [a]]`. Each channel is a number
    between 0-255 and `a` is 255 if not supplied.

    default [0, 0, 0, 255]
    """

    font_settings = t.Any(None, allow_none=True).tag(sync=True)
    """
    Advance options for fine tuning the appearance and performance of the generated
    shared `fontAtlas`.
    """

    word_break = t.Any(None, allow_none=True).tag(sync=True)
    """
    Available options are `break-all` and `break-word`. A valid `maxWidth` has to be
    provided to use `wordBreak`.

    default 'break-word'
    """

    max_width = t.Any(None, allow_none=True).tag(sync=True)
    """
    A unitless number that will be multiplied with the current text size to set the
    width limit of a string.

    If specified, when the text is longer than the width limit, it will be wrapped into
    multiple lines using the strategy of `wordBreak`.

    default -1
    """

    get_text = TextAccessor(None, allow_none=True)
    """Label text accessor"""

    # get_position = t.Any(None, allow_none=True).tag(sync=True)
    # """Anchor position accessor"""

    #  ?: Accessor<DataT, Position>;

    get_color = ColorAccessor(None, allow_none=True)
    """Label color accessor

    default [0, 0, 0, 255]
    """

    get_size = FloatAccessor(None, allow_none=True)
    """Label size accessor

    default 32
    """

    get_angle = FloatAccessor(None, allow_none=True)
    """Label rotation accessor, in degrees

    default 0
    """

    get_text_anchor = t.Any(None, allow_none=True).tag(sync=True)
    """Horizontal alignment accessor

    default 'middle'
    """
    #  ?: Accessor<DataT, 'start' | 'middle' | 'end'>;

    get_alignment_baseline = t.Any(None, allow_none=True).tag(sync=True)
    """Vertical alignment accessor

    default 'center'
    """
    #  ?: Accessor<DataT, 'top' | 'center' | 'bottom'>;

    get_pixel_offset = t.Any(None, allow_none=True).tag(sync=True)
    """Label offset from the anchor position, [x, y] in pixels

    default [0, 0]
    """
    #  ?: Accessor<DataT, [number, number]>;


class TripsLayer(BaseArrowLayer):
    """The `TripsLayer` renders animated paths that represent moving objects.

    The easiest way to create a `TripsLayer` is by using the
    [`from_movingpandas`][lonboard.experimental.TripsLayer.from_movingpandas]
    constructor, where you can pass in a [`movingpandas.TrajectoryCollection`][].

    Otherwise, this layer requires a specific Arrow input data schema to associate the
    timestamp of each coordinate with the spatial information in the `LineString`
    geometries.

    In order to animate this layer, call the
    [`animate`][lonboard.experimental.TripsLayer.animate] method.

    !!! warning

        The TripsLayer renders data representing a specific instance in time based on
        the [`current_time`][lonboard.experimental.TripsLayer.current_time] attribute.

        If you don't see any data, use the
        [`animate`][lonboard.experimental.TripsLayer.animate] method to automatically
        set `current_time`.

    !!! info "Passing in custom Arrow data"

        As with all layers, you can pass an Arrow `Table` into the `table` parameter of
        this layer. In the case of the `TripsLayer`, there must be one column in the
        table that is a GeoArrow `LineString` geometry column, tagged with the
        `geoarrow.linestring` extension name. Additionally, the
        [`get_timestamps`][lonboard.experimental.TripsLayer.get_timestamps] accessor
        needs to be an Arrow list-typed column with a timestamp-typed child array. The
        `get_timestamps` column must have the **exact same nesting** as the `LineString`
        column. That is, there must be one timestamp for every coordinate in the
        `LineString` column. This is validated in data input.
    """

    _animation_link: ipywidgets.widgets.widget_link.DirectionalLink | None = None
    """
    An ipywidgets link created and managed by `animate`. A reference is stored here so
    that we can call `unlink` on the previous link object if `animate` is called
    multiple times.
    """

    _layer_type = traitlets.Unicode("trip").tag(sync=True)

    table = ArrowTableTrait(
        allowed_geometry_types={
            EXTENSION_NAME.LINESTRING,
        },
    )

    width_units = traitlets.Unicode(default_value=None, allow_none=True).tag(sync=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    width_scale = traitlets.Float(default_value=None, allow_none=True, min=0).tag(
        sync=True,
    )
    """
    The path width multiplier that multiplied to all paths.

    - Type: `float`, optional
    - Default: `1`
    """

    width_min_pixels = traitlets.Float(default_value=None, allow_none=True, min=0).tag(
        sync=True,
    )
    """
    The minimum path width in pixels. This prop can be used to prevent the path from
    getting too thin when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    width_max_pixels = traitlets.Float(default_value=None, allow_none=True, min=0).tag(
        sync=True,
    )
    """
    The maximum path width in pixels. This prop can be used to prevent the path from
    getting too thick when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    joint_rounded = traitlets.Bool(default_value=None, allow_none=True).tag(sync=True)
    """
    Type of joint. If `True`, draw round joints. Otherwise draw miter joints.

    - Type: `bool`, optional
    - Default: `False`
    """

    cap_rounded = traitlets.Bool(default_value=None, allow_none=True).tag(sync=True)
    """
    Type of caps. If `True`, draw round caps. Otherwise draw square caps.

    - Type: `bool`, optional
    - Default: `False`
    """

    miter_limit = traitlets.Int(default_value=None, allow_none=True).tag(sync=True)
    """
    The maximum extent of a joint in ratio to the stroke width.
    Only works if `jointRounded` is `False`.

    - Type: `float`, optional
    - Default: `4`
    """

    billboard = traitlets.Bool(default_value=None, allow_none=True).tag(sync=True)
    """
    If `True`, extrude the path in screen space (width always faces the camera).
    If `False`, the width always faces up.

    - Type: `bool`, optional
    - Default: `False`
    """

    fade_trail = traitlets.Bool(default_value=None, allow_none=True).tag(sync=True)
    """Whether or not the path fades out.

    - Type: `bool`, optional
    - Default: `True`
    """

    trail_length = traitlets.Float(default_value=None, allow_none=True).tag(sync=True)
    """Trail length.

    - Type: `float`, optional
    - Default: `120`
    """

    _current_time = traitlets.Float(0).tag(sync=True)
    """The current time of the frame.

    - Type: `float`, optional
    - Default: set by the [`animate`][lonboard.experimental.TripsLayer.animate] method.

    !!! info
        This `_current_time` is not directly interpretable.

        Because of some technical details in deck.gl (the fact that deck.gl represents
        timestamps as `float32`), the `TripsLayer` automatically rescales the input
        timestamp data to a range representable by `float32`.
    """

    get_color = ColorAccessor(None, allow_none=True)
    """
    The color of each path in the format of `[r, g, b, [a]]`. Each channel is a number
    between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the color for all
          paths.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the color for the path at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_width = FloatAccessor(None, allow_none=True)
    """
    The width of each path, in units specified by `width_units` (default `'meters'`).

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all paths.
        - If an array is provided, each value in the array will be used as the width for
          the path at the same row index.
    - Default: `1`.
    """

    get_timestamps = TimestampAccessor(None, allow_none=True)
    """
    The timestamp of each coordinate.

    - Type: [TimestampAccessor][lonboard.experimental.traits.TimestampAccessor]
    """

    def __init__(
        self,
        *,
        table: ArrowStreamExportable,
        get_timestamps: ArrowStreamExportable,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[TripsLayerKwargs],
    ) -> None:
        """Construct a TripsLayer from existing Arrow data.

        Args:
            table: An Arrow table from any Arrow-compatible library with a
                `geoarrow.linestring` geometry column.
            get_timestamps: An Arrow column or ChunkedArray with timestamps that
                correspond to the `LineString` geometries in the main table. This column
                must be a list-typed array containing a timestamp-typed child array.

        """
        super().__init__(
            table=table,
            _rows_per_chunk=_rows_per_chunk,
            get_timestamps=get_timestamps,  # type: ignore
            **kwargs,
        )

    @classmethod
    def from_geopandas(  # type: ignore
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        get_timestamps: ArrowStreamExportable,
        auto_downcast: bool = True,
        **kwargs: Unpack[TripsLayerKwargs],
    ) -> Self:
        return super().from_geopandas(
            gdf=gdf,
            auto_downcast=auto_downcast,
            get_timestamps=get_timestamps,  # type: ignore
            **kwargs,
        )

    @classmethod
    def from_duckdb(  # type: ignore
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        get_timestamps: ArrowStreamExportable,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[TripsLayerKwargs],
    ) -> Self:
        return super().from_duckdb(
            sql=sql,
            con=con,
            crs=crs,
            get_timestamps=get_timestamps,  # type: ignore
            **kwargs,
        )

    @classmethod
    def from_movingpandas(
        cls,
        traj_collection: TrajectoryCollection,
        **kwargs: Unpack[TripsLayerKwargs],
    ) -> Self:
        """Construct a TripsLayer from a MovingPandas
        [`TrajectoryCollection`][movingpandas.TrajectoryCollection].

        This is the simplest way to construct a `TripsLayer`. Under the hood, this
        constructor will convert the `TrajectoryCollection` to a GeoArrow table with a
        `LineString` geometry column for each row. It will also create a Timestamp Arrow
        column from the TrajectoryCollection's [`DatetimeIndex`][pandas.DatetimeIndex].

        Args:
            traj_collection: the trajectory collection

        Keyword Args:
            kwargs: keyword args to pass to the `TripsLayer` constructor.

        """
        from lonboard._geoarrow.movingpandas_interop import movingpandas_to_geoarrow

        (table, timestamp_col) = movingpandas_to_geoarrow(
            traj_collection=traj_collection,
        )
        return cls(table=table, get_timestamps=timestamp_col, **kwargs)

    def animate(
        self,
        *,
        step: timedelta,
        fps: float = 50,
        strftime_fmt: str | None = None,
    ) -> ipywidgets.widgets.widget_box.HBox:
        """Animate this layer with an
        [`ipywidgets.Play`][ipywidgets.widgets.widget_int.Play] controller.

        You can change how "fast" the animation is perceived by either increasing the
        amount of "data time" in each animation step, or by having more animation frames
        per second.

        As an example, passing `step=timedelta(seconds=60)` will set each time step of
        the animation to be 60 "data seconds". Setting `fps=50` (the default) causes
        there to be 50 animation frames per second.

        Note that for large data, it's possible there will be some rendering lag and
        data may not actually update at the desired frames per second.

        If you call `animate` multiple times, only the most recently produced `Play`
        widget will be active and linked to the map.

        Keyword Args:
            step: the length of time in the data to progress between each animation
                frame.
            fps: the number of animation frames per second. Defaults to `50`.
            strftime_fmt: the format string passed to
                [`datetime.strftime`][datetime.datetime.strftime]. Defaults to an
                inferred format string.

        Returns:
            an [`ipywidgets.HBox`][ipywidgets.widgets.widget_box.HBox] containing an
                [`ipywidgets.Play`][ipywidgets.widgets.widget_int.Play] controller to
                manage animation playback and an
                [`ipywidgets.Output`][ipywidgets.widgets.widget_output.Output] to
                display the current datetime of the map.

        """
        assert isinstance(step, timedelta), "expected step to be a timedelta."

        interval = 1000 / fps

        time_unit = self.get_timestamps.type.value_type.time_unit
        assert time_unit is not None

        # Convert the `step` into the units used by the timestamps on the layer.
        if time_unit == "s":
            resolved_step = step.total_seconds()
        elif time_unit == "ms":
            resolved_step = step.total_seconds() * 1000
        elif time_unit == "us":
            resolved_step = step.total_seconds() * 1000**2
        elif time_unit == "ns":
            resolved_step = step.total_seconds() * 1000**3
        else:
            assert False, f"Unexpected time unit {time_unit}"

        # Unlink any previous animation widget
        self.stop_animation()

        max_value = timestamp_max_physical_value(self.get_timestamps)
        play_widget = ipywidgets.Play(
            value=MIN_INTEGER_FLOAT32,
            min=MIN_INTEGER_FLOAT32,
            max=max_value,
            step=resolved_step,
            interval=interval,
            repeat=True,
        )

        timestamp_output = ipywidgets.Output()

        # Store a reference to the widget link
        self._animation_link = ipywidgets.jsdlink(
            (play_widget, "value"),
            (self, "_current_time"),
        )

        # Some callback prep done outside of the callback for perf
        tz_str: str | None = self.get_timestamps.type.value_type.tz

        # Custom datetime format without Z/local time because we append the tz string
        # below
        default_strftime = "%Y-%m-%d %H:%M:%S"
        if time_unit != "s":
            default_strftime += ".%f"

        def update_timestamp_output(change) -> None:
            current_datetime = self._current_time_to_datetime(change["new"])
            # Ignore timezone info because we also print tz str
            current_datetime.replace(tzinfo=None)

            if strftime_fmt is not None:
                s = current_datetime.strftime(strftime_fmt)
            else:
                s = current_datetime.strftime(default_strftime)

            if tz_str is not None:
                s += f" {tz_str}"

            timestamp_output.clear_output(wait=True)
            with timestamp_output:
                print(s)

        play_widget.observe(update_timestamp_output, names="value")

        return ipywidgets.HBox([play_widget, timestamp_output])

    @property
    def current_time(self) -> datetime:
        """Get the current time of the map as a `datetime` object.

        Returns:
            datetime object with current time.

        """
        return self._current_time_to_datetime(self._current_time)

    def _current_time_to_datetime(self, current_time: float) -> datetime:
        start_offset = timestamp_start_offset(self.get_timestamps)
        timestamp_int = int(current_time - start_offset)
        timestamp_scalar = Scalar(timestamp_int, type=DataType.int64()).cast(
            self.get_timestamps.type.value_type,
        )
        return timestamp_scalar.as_py()

    def stop_animation(self):
        """Stop any existing animation.

        This will
        [unlink](https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20Basics.html#unlinking-widgets)
        the linking between the `Play` widget generated by `animate` and the layer.

        To reanimate the map, call [`animate`][lonboard.experimental.TripsLayer.animate]
        again.
        """
        if self._animation_link is not None:
            self._animation_link.unlink()
            self._animation_link = None



"""Compute the total bounds of a geoarrow column."""

from __future__ import annotations

import math
from dataclasses import dataclass

import numpy as np
from arro3.core import Array, ChunkedArray, DataType, Field, list_flatten, struct_field

from lonboard._constants import EXTENSION_NAME


@dataclass
class Bbox:
    minx: float = math.inf
    miny: float = math.inf
    maxx: float = -math.inf
    maxy: float = -math.inf

    def update(self, other: Bbox) -> None:
        self.minx = min(self.minx, other.minx)
        self.miny = min(self.miny, other.miny)
        self.maxx = max(self.maxx, other.maxx)
        self.maxy = max(self.maxy, other.maxy)

    def to_tuple(self) -> tuple[float, float, float, float]:
        return (self.minx, self.miny, self.maxx, self.maxy)


def total_bounds(field: Field, column: ChunkedArray) -> Bbox:
    """Compute the total bounds of a geometry column."""
    extension_type_name = field.metadata[b"ARROW:extension:name"]

    if extension_type_name == EXTENSION_NAME.POINT:
        return _total_bounds_nest_0(column)

    if extension_type_name in [EXTENSION_NAME.LINESTRING, EXTENSION_NAME.MULTIPOINT]:
        return _total_bounds_nest_1(column)

    if extension_type_name in [EXTENSION_NAME.POLYGON, EXTENSION_NAME.MULTILINESTRING]:
        return _total_bounds_nest_2(column)

    if extension_type_name == EXTENSION_NAME.MULTIPOLYGON:
        return _total_bounds_nest_3(column)

    if extension_type_name == EXTENSION_NAME.BOX:
        return _total_bounds_box(column)

    assert False


def _coords_bbox(arr: Array) -> Bbox:
    assert DataType.is_fixed_size_list(arr.type)
    list_size = arr.type.list_size
    assert list_size is not None

    np_arr = list_flatten(arr).to_numpy().reshape(-1, list_size)
    min_vals = np.min(np_arr, axis=0)
    max_vals = np.max(np_arr, axis=0)
    return Bbox(minx=min_vals[0], miny=min_vals[1], maxx=max_vals[0], maxy=max_vals[1])


def _total_bounds_nest_0(column: ChunkedArray) -> Bbox:
    bbox = Bbox()
    for coords in column.chunks:
        bbox.update(_coords_bbox(coords))

    return bbox


def _total_bounds_nest_1(column: ChunkedArray) -> Bbox:
    bbox = Bbox()
    flat_array = list_flatten(column)
    for coords in flat_array:
        bbox.update(_coords_bbox(coords))

    return bbox


def _total_bounds_nest_2(column: ChunkedArray) -> Bbox:
    bbox = Bbox()
    flat_array = list_flatten(list_flatten(column))
    for coords in flat_array:
        bbox.update(_coords_bbox(coords))

    return bbox


def _total_bounds_nest_3(column: ChunkedArray) -> Bbox:
    bbox = Bbox()
    flat_array = list_flatten(list_flatten(list_flatten(column)))
    for coords in flat_array:
        bbox.update(_coords_bbox(coords))

    return bbox


def _total_bounds_box(column: ChunkedArray) -> Bbox:
    """Compute the total bounds of a geoarrow.box column."""
    bbox = Bbox()
    for chunk in column.chunks:
        is_2d = len(chunk.field.type.fields) == 4
        is_3d = len(chunk.field.type.fields) == 6

        if is_2d:
            minx = np.min(struct_field(chunk, 0))
            miny = np.min(struct_field(chunk, 1))
            maxx = np.max(struct_field(chunk, 2))
            maxy = np.max(struct_field(chunk, 3))
        elif is_3d:
            minx = np.min(struct_field(chunk, 0))
            miny = np.min(struct_field(chunk, 1))
            maxx = np.max(struct_field(chunk, 3))
            maxy = np.max(struct_field(chunk, 4))
        else:
            raise ValueError(
                f"Unexpected box type with {len(chunk.field.type.fields)} fields.\n"
                "Only 2D and 3D boxes are supported.",
            )
        bbox.update(Bbox(minx=minx, miny=miny, maxx=maxx, maxy=maxy))

    return bbox



import numpy as np
from arro3.core import Array, ChunkedArray, Table


def add_positional_row_index(
    table: Table,
) -> Table:
    num_rows = table.num_rows
    if num_rows <= np.iinfo(np.uint8).max:
        arange_col = Array(np.arange(num_rows, dtype=np.uint8))
    elif num_rows <= np.iinfo(np.uint16).max:
        arange_col = Array(np.arange(num_rows, dtype=np.uint16))
    elif num_rows <= np.iinfo(np.uint32).max:
        arange_col = Array(np.arange(num_rows, dtype=np.uint32))
    else:
        arange_col = Array(np.arange(num_rows, dtype=np.uint64))

    return table.append_column("row_index", ChunkedArray([arange_col]))



"""Reproject a GeoArrow array."""

# ruff: noqa: RET504

from __future__ import annotations

import json
import warnings
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache, partial
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
from arro3.core import (
    Array,
    ChunkedArray,
    DataType,
    Field,
    Table,
    fixed_size_list_array,
    list_array,
    list_flatten,
    list_offsets,
)
from pyproj import CRS, Transformer

from lonboard._constants import EPSG_4326, EXTENSION_NAME, OGC_84
from lonboard._geoarrow.crs import get_field_crs
from lonboard._geoarrow.extension_types import CoordinateDimension
from lonboard._utils import get_geometry_column_index

TransformerFromCRS = lru_cache(Transformer.from_crs)

if TYPE_CHECKING:
    from collections.abc import Callable


def no_crs_warning() -> None:
    warn(
        "No CRS exists on data. "
        "If no data is shown on the map, double check that your CRS is WGS84.",
    )


def reproject_table(
    table: Table,
    *,
    to_crs: str | CRS = OGC_84,
    max_workers: int | None = None,
) -> Table:
    """Reproject a GeoArrow table to a new CRS.

    Args:
        table: The table to reproject.
        to_crs: The target CRS. Defaults to OGC_84.
        max_workers: The maximum number of threads to use. Defaults to None.

    Returns:
        A new table.

    """
    geom_col_idx = get_geometry_column_index(table.schema)
    # No geometry column in table
    if geom_col_idx is None:
        return table

    geom_field = table.schema.field(geom_col_idx)
    geom_column = table.column(geom_col_idx)

    # geometry column exists in table but is not assigned a CRS
    if b"ARROW:extension:metadata" not in geom_field.metadata:
        no_crs_warning()
        return table

    new_field, new_column = reproject_column(
        field=geom_field,
        column=geom_column,
        to_crs=to_crs,
        max_workers=max_workers,
    )
    return table.set_column(geom_col_idx, new_field, new_column)


def reproject_column(
    *,
    field: Field,
    column: ChunkedArray,
    to_crs: str | CRS = OGC_84,
    max_workers: int | None = None,
) -> tuple[Field, ChunkedArray]:
    """Reproject a GeoArrow array to a new CRS.

    Args:
        field: The field describing the column
        column: A ChunkedArray
        to_crs: The target CRS. Defaults to OGC_84.
        max_workers: The maximum number of threads to use. Defaults to None.

    """
    extension_type_name = field.metadata[b"ARROW:extension:name"]
    existing_crs = get_field_crs(field)
    if existing_crs is None:
        no_crs_warning()
        return field, column

    if existing_crs == to_crs:
        return field, column

    # If projecting to OGC_84, also check if existing CRS is EPSG_4326, which when
    # passing always_xy is equivalent.
    if to_crs == OGC_84 and existing_crs == EPSG_4326:
        return field, column

    # NOTE: Not sure the best place to put this warning
    warnings.warn(
        "Input being reprojected to EPSG:4326 CRS.\n"
        "Lonboard is only able to render data in EPSG:4326 projection.",
    )

    transformer = TransformerFromCRS(existing_crs, to_crs, always_xy=True)

    # Metadata inside metadata, bad naming
    new_extension_meta_meta = {"crs": CRS(to_crs).to_json()}
    new_extension_metadata = {
        b"ARROW:extension:name": extension_type_name,
        b"ARROW:extension:metadata": json.dumps(new_extension_meta_meta).encode(),
    }

    new_chunked_array = _reproject_column(
        column,
        extension_type_name=extension_type_name,  # type: ignore
        transformer=transformer,
        max_workers=max_workers,
    )
    new_field = field.with_type(new_chunked_array.type).with_metadata(
        new_extension_metadata,
    )
    return new_field, new_chunked_array


def _reproject_column(
    column: ChunkedArray,
    *,
    extension_type_name: EXTENSION_NAME,
    transformer: Transformer,
    max_workers: int | None = None,
) -> ChunkedArray:
    if extension_type_name == EXTENSION_NAME.POINT:
        func = partial(_reproject_chunk_nest_0, transformer=transformer)
    elif extension_type_name in [EXTENSION_NAME.LINESTRING, EXTENSION_NAME.MULTIPOINT]:
        func = partial(_reproject_chunk_nest_1, transformer=transformer)
    elif extension_type_name in [
        EXTENSION_NAME.POLYGON,
        EXTENSION_NAME.MULTILINESTRING,
    ]:
        func = partial(_reproject_chunk_nest_2, transformer=transformer)

    elif extension_type_name == EXTENSION_NAME.MULTIPOLYGON:
        func = partial(_reproject_chunk_nest_3, transformer=transformer)
    else:
        raise ValueError(f"Unexpected extension type name {extension_type_name}")

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        return ChunkedArray(list(executor.map(func, column.chunks)))


def _reproject_coords(arr: Array, transformer: Transformer) -> Array:
    list_size = arr.type.list_size
    assert list_size is not None
    np_arr = list_flatten(arr).to_numpy().reshape(-1, list_size)

    if list_size == 2:
        output_np_arr = np.column_stack(
            transformer.transform(np_arr[:, 0], np_arr[:, 1]),
        )
        dims = CoordinateDimension.XY
    elif list_size == 3:
        output_np_arr = np.column_stack(
            transformer.transform(np_arr[:, 0], np_arr[:, 1], np_arr[:, 2]),
        )
        dims = CoordinateDimension.XYZ
    else:
        raise ValueError(f"Unexpected list size {list_size}")

    coord_field = DataType.list(Field(dims, DataType.float64()), len(dims))
    return fixed_size_list_array(output_np_arr.ravel("C"), len(dims), type=coord_field)


def _reproject_chunk_nest_0(arr: Array, transformer: Transformer) -> Array:
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_0(arr, callback)


def _reproject_chunk_nest_1(arr: Array, transformer: Transformer) -> Array:
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_1(arr, callback)


def _reproject_chunk_nest_2(arr: Array, transformer: Transformer) -> Array:
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_2(arr, callback)


def _reproject_chunk_nest_3(arr: Array, transformer: Transformer) -> Array:
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_3(arr, callback)


def _map_coords_nest_0(
    arr: Array,
    callback: Callable[[Array], Array],
) -> Array:
    new_coords = callback(arr)
    return new_coords


def _map_coords_nest_1(
    arr: Array,
    callback: Callable[[Array], Array],
) -> Array:
    geom_offsets = list_offsets(arr, logical=True)
    coords = list_flatten(arr)
    new_coords = callback(coords)
    new_geometry_array = list_array(geom_offsets, new_coords)
    return new_geometry_array


def _map_coords_nest_2(
    arr: Array,
    callback: Callable[[Array], Array],
) -> Array:
    geom_offsets = list_offsets(arr, logical=True)
    ring_offsets = list_offsets(list_flatten(arr), logical=True)
    coords = list_flatten(list_flatten(arr))
    new_coords = callback(coords)
    new_ring_array = list_array(ring_offsets, new_coords)
    new_geometry_array = list_array(geom_offsets, new_ring_array)
    return new_geometry_array


def _map_coords_nest_3(
    arr: Array,
    callback: Callable[[Array], Array],
) -> Array:
    geom_offsets = list_offsets(arr, logical=True)
    polygon_offsets = list_offsets(list_flatten(arr), logical=True)
    ring_offsets = list_offsets(list_flatten(list_flatten(arr)), logical=True)
    coords = list_flatten(list_flatten(list_flatten(arr)))
    new_coords = callback(coords)
    new_ring_array = list_array(ring_offsets, new_coords)
    new_polygon_array = list_array(polygon_offsets, new_ring_array)
    new_geometry_array = list_array(geom_offsets, new_polygon_array)
    return new_geometry_array



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from pyproj import CRS

if TYPE_CHECKING:
    from arro3.core import Field


# Note: According to the spec, if the metadata key exists, its value should never be
# `null` or an empty dict, but we still check for those to be safe
def get_field_crs(field: Field) -> CRS | None:
    extension_metadata_value = field.metadata_str.get("ARROW:extension:metadata")
    if not extension_metadata_value:
        return None

    extension_metadata = json.loads(extension_metadata_value)
    return parse_metadata(extension_metadata)


def parse_metadata(extension_metadata: dict[str, Any]) -> CRS | None:
    crs_val = extension_metadata.get("crs")
    crs_type = extension_metadata.get("crs_type")

    if crs_type == "projjson":
        assert crs_val is not None, "CRS value must be provided for projjson type"
        return CRS.from_json_dict(crs_val)

    if crs_type == "wkt2:2019":
        assert crs_val is not None, "CRS value must be provided for WKT2:2019 type"
        return CRS.from_wkt(crs_val)

    if crs_type == "authority_code":
        assert crs_val is not None, "CRS value must be provided for authority code type"
        assert ":" in crs_val, "Authority code must be in the format 'authority:code'"
        assert isinstance(crs_val, str), "CRS value must be a string"
        return CRS.from_authority(*crs_val.split(":", 1))

    return CRS.from_user_input(crs_val) if crs_val is not None else None


def serialize_crs(crs: CRS | None) -> dict[str, Any] | None:
    """Serialize a CRS to GeoArrow metadata.

    The returned GeoArrow metadata should be JSON-serialized within the
    ARROW:extension:metadata key.
    """
    if crs is None:
        return None

    return {
        "crs": crs.to_json_dict(),
        "crs_type": "projjson",
    }



"""Python library for fast, interactive geospatial vector data visualization in Jupyter."""

from . import colormap, controls, experimental, layer_extension, traits
from ._layer import (
    BaseArrowLayer,
    BaseLayer,
    BitmapLayer,
    BitmapTileLayer,
    ColumnLayer,
    HeatmapLayer,
    PathLayer,
    PointCloudLayer,
    PolygonLayer,
    ScatterplotLayer,
    SolidPolygonLayer,
)
from ._map import Map
from ._version import __version__
from ._viz import viz



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import numpy as np
from arro3.core import (
    Array,
    ChunkedArray,
    DataType,
    Field,
    RecordBatch,
    Schema,
    Table,
    fixed_size_list_array,
    list_array,
)

if TYPE_CHECKING:
    import movingpandas as mpd
    import pandas as pd
    import pyarrow as pa
    from movingpandas import TrajectoryCollection


def movingpandas_to_geoarrow(  # noqa: PLR0915
    traj_collection: TrajectoryCollection,
) -> tuple[Table, ChunkedArray]:
    """Convert a MovingPandas TrajectoryCollection to GeoArrow.

    Args:
        traj_collection: _description_

    Returns:
        _description_

    """
    import pyarrow as pa
    import shapely

    crs = traj_collection.get_crs()
    crs_json = crs.to_json_dict() if crs is not None else None

    num_coords = 0
    num_trajectories = len(traj_collection)
    offsets = np.zeros(num_trajectories + 1, dtype=np.int32)
    datetime_dtypes = set()
    attr_schemas: list[pa.Schema] = []

    # Loop the first time to infer offsets for each trajectory
    for i, traj in enumerate(traj_collection.trajectories):
        traj: mpd.Trajectory

        num_coords += traj.size()
        offsets[i + 1] = num_coords
        datetime_dtypes.add(traj.df.index.dtype)

        geom_col_name = traj.get_geom_col()
        df_attr = traj.df.drop(columns=[geom_col_name])

        # Explicitly drop index because the index is a DatetimeIndex that we convert
        # manually later.
        arrow_schema = pa.Schema.from_pandas(df_attr, preserve_index=False)
        attr_schemas.append(arrow_schema)

    assert len(datetime_dtypes) == 1, (
        "Expected only one datetime dtype across all trajectories."
    )
    datetime_dtype = list(datetime_dtypes)[0]

    # We currently always provision space for XYZ coordinates, and then only use 2d if
    # the Z dimension is always NaN
    coords = np.zeros((num_coords, 3), dtype=np.float64)

    arrow_timestamp_dtype = infer_timestamp_dtype(datetime_dtype)
    timestamps = np.zeros(num_coords, dtype=np.int64)

    attr_schema = pa.unify_schemas(attr_schemas, promote_options="permissive")
    attr_tables: list[pa.Table] = []

    # Loop second time to fill timestamps and coords
    for i, traj in enumerate(traj_collection.trajectories):
        start_offset = offsets[i]
        end_offset = offsets[i + 1]

        timestamps[start_offset:end_offset] = traj.df.index
        coords[start_offset:end_offset] = shapely.get_coordinates(
            traj.df.geometry.array,  # type: ignore
            include_z=True,
        )

        geom_col_name = traj.get_geom_col()
        df_attr = traj.df.drop(columns=[geom_col_name])

        attr_table = pa.Table.from_pandas(
            traj.df,
            schema=attr_schema,
            preserve_index=False,
        )
        attr_tables.append(attr_table)

    attr_table = pa.concat_tables(attr_tables, promote_options="none")
    attr_table = Table.from_arrow(attr_table)

    offsets = Array.from_numpy(offsets)

    nested_attr_table = apply_offsets_to_table(attr_table, offsets=offsets)

    if np.all(np.isnan(coords[:, 2])):
        coord_list_size = 2
        # Cast to 2D coords
        coords = coords[:, :2]
    else:
        assert not np.any(
            np.isnan(coords[:, 2]),
        ), "Mixed 2D and 3D coordinates not currently supported"
        coord_list_size = 3

    coords_arr = Array.from_numpy(coords.ravel("C"))
    coords_fixed_size_list = fixed_size_list_array(coords_arr, coord_list_size)
    linestrings_arr = list_array(offsets, coords_fixed_size_list)

    extension_metadata: dict[str, str] = {"ARROW:extension:name": "geoarrow.linestring"}
    if crs_json is not None:
        extension_metadata["ARROW:extension:metadata"] = json.dumps({"crs": crs_json})

    linestrings_field = Field(
        "geometry",
        linestrings_arr.type,
        nullable=True,
        metadata=extension_metadata,
    )

    timestamp_values = Array.from_numpy(timestamps).cast(arrow_timestamp_dtype)
    timestamp_arr = list_array(offsets, timestamp_values)
    timestamp_col = ChunkedArray([timestamp_arr])

    table = nested_attr_table.append_column(
        linestrings_field,
        ChunkedArray([linestrings_arr]),
    )
    return table, timestamp_col


def infer_timestamp_dtype(dtype: np.dtype | pd.DatetimeTZDtype) -> DataType:
    """Infer an arrow time unit from the numpy data type.

    Raises:
        ValueError: If not a known numpy datetime dtype

    """
    import pandas as pd

    if isinstance(dtype, pd.DatetimeTZDtype):
        return DataType.timestamp(dtype.unit, tz=str(dtype.tz))

    if dtype.name == "datetime64[s]":
        return DataType.timestamp("s")

    if dtype.name == "datetime64[ms]":
        return DataType.timestamp("ms")

    if dtype.name == "datetime64[us]":
        return DataType.timestamp("us")

    if dtype.name == "datetime64[ns]":
        return DataType.timestamp("ns")

    raise ValueError(f"Unexpected datetime type: {dtype}")


def apply_offsets_to_table(table: Table, offsets: Array) -> Table:
    batch = table.combine_chunks().to_batches()[0]

    new_fields = []
    new_arrays = []

    for field_idx in range(batch.num_columns):
        field = batch.schema.field(field_idx)
        new_field = field.with_type(DataType.list(field))
        new_array = list_array(offsets, batch[field_idx], type=new_field)

        new_fields.append(new_field)
        new_arrays.append(new_array)

    new_schema = Schema(new_fields, metadata=batch.schema.metadata)
    new_batch = RecordBatch(new_arrays, schema=new_schema)
    return Table.from_batches([new_batch])



"""Basemap helpers."""

from enum import Enum


class CartoBasemap(str, Enum):
    """Basemap styles provided by Carto.

    Refer to [Carto
    documentation](https://docs.carto.com/carto-for-developers/carto-for-react/guides/basemaps)
    for information on styles.
    """

    DarkMatter = "https://basemaps.cartocdn.com/gl/dark-matter-gl-style/style.json"
    """A dark map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/dark_labels.png)
    """

    DarkMatterNoLabels = (
        "https://basemaps.cartocdn.com/gl/dark-matter-nolabels-gl-style/style.json"
    )
    """A dark map style without labels."""

    Positron = "https://basemaps.cartocdn.com/gl/positron-gl-style/style.json"
    """A light map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/positron_labels.png)
    """

    PositronNoLabels = (
        "https://basemaps.cartocdn.com/gl/positron-nolabels-gl-style/style.json"
    )
    """A light map style without labels."""

    Voyager = "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json"
    """A light, colored map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/voyager_labels.png)
    """

    VoyagerNoLabels = (
        "https://basemaps.cartocdn.com/gl/voyager-nolabels-gl-style/style.json"
    )
    """A light, colored map style without labels."""



from __future__ import annotations

import math
from io import BytesIO
from typing import TYPE_CHECKING, Any, overload

import arro3.compute as ac
from arro3.core import (
    Array,
    ChunkedArray,
    DataType,
    RecordBatch,
    Scalar,
    Table,
    list_array,
    list_flatten,
    list_offsets,
)
from traitlets import TraitError

from lonboard._utils import timestamp_start_offset

if TYPE_CHECKING:
    from lonboard._layer import BaseArrowLayer
    from lonboard.experimental._layer import TripsLayer
    from lonboard.models import ViewState


DEFAULT_PARQUET_COMPRESSION = "ZSTD"
DEFAULT_PARQUET_COMPRESSION_LEVEL = 7
DEFAULT_PARQUET_CHUNK_SIZE = 2**16
# Target chunk size for Arrow (uncompressed) per Parquet chunk
DEFAULT_ARROW_CHUNK_BYTES_SIZE = 5 * 1024 * 1024  # 5MB

# Maximum number of separate chunks/row groups to allow splitting an input layer into
# Deck.gl can pick from a maximum of 256 layers, and a user could have many layers, so
# we don't want to use too many layers per data file.
DEFAULT_MAX_NUM_CHUNKS = 32


def write_parquet_batch(record_batch: RecordBatch) -> bytes:
    """Write a RecordBatch to a Parquet file.

    We still use pyarrow.parquet.ParquetWriter if pyarrow is installed because pyarrow
    has better encoding defaults. So Parquet files written by pyarrow are smaller by
    default than files written by arro3.io.write_parquet.
    """
    # Occasionally it's possible for there to be empty batches in the
    # pyarrow table. This will error when writing to parquet. We want to
    # give a more informative error.
    if record_batch.num_rows == 0:
        raise ValueError("Batch with 0 rows.")

    try:
        import pyarrow as pa
        import pyarrow.parquet as pq

        bio = BytesIO()
        with pq.ParquetWriter(
            bio,
            schema=pa.schema(record_batch.schema),
            compression=DEFAULT_PARQUET_COMPRESSION,
            compression_level=DEFAULT_PARQUET_COMPRESSION_LEVEL,
        ) as writer:
            writer.write_batch(
                pa.record_batch(record_batch),
                row_group_size=record_batch.num_rows,
            )

        return bio.getvalue()

    except ImportError:
        from arro3.io import write_parquet

        compression_string = (
            f"{DEFAULT_PARQUET_COMPRESSION}({DEFAULT_PARQUET_COMPRESSION_LEVEL})"
        )
        bio = BytesIO()
        write_parquet(
            record_batch,
            bio,
            compression=compression_string,
            max_row_group_size=record_batch.num_rows,
        )

        return bio.getvalue()


def serialize_table_to_parquet(table: Table, *, max_chunksize: int) -> list[bytes]:
    buffers: list[bytes] = []
    assert max_chunksize > 0

    for record_batch in table.rechunk(max_chunksize=max_chunksize).to_batches():
        buffers.append(write_parquet_batch(record_batch))

    return buffers


def serialize_pyarrow_column(
    data: Array | ChunkedArray,
    *,
    max_chunksize: int,
) -> list[bytes]:
    """Serialize a pyarrow column to a Parquet file with one column."""
    pyarrow_table = Table.from_pydict({"value": data})
    return serialize_table_to_parquet(pyarrow_table, max_chunksize=max_chunksize)


@overload
def serialize_accessor(
    data: ChunkedArray,
    obj: BaseArrowLayer,
) -> list[bytes]: ...
@overload
def serialize_accessor(
    data: str | float | list | tuple | bytes,
    obj: BaseArrowLayer,
) -> str | int | float | list | tuple | bytes: ...
def serialize_accessor(
    data: str | float | list | tuple | bytes | ChunkedArray,
    obj: BaseArrowLayer,
):
    if data is None:
        return None

    # We assume data has already been validated to the right type for this accessor
    # Allow any json-serializable type through
    if isinstance(data, (str, int, float, list, tuple, bytes)):
        return data

    assert isinstance(data, ChunkedArray)
    validate_accessor_length_matches_table(data, obj.table)
    return serialize_pyarrow_column(data, max_chunksize=obj._rows_per_chunk)  # noqa: SLF001


def serialize_table(data: Table, obj: BaseArrowLayer) -> list[bytes]:
    assert isinstance(data, Table), "expected Arrow table"
    return serialize_table_to_parquet(data, max_chunksize=obj._rows_per_chunk)  # noqa: SLF001


def infer_rows_per_chunk(table: Table) -> int:
    # At least one chunk
    num_chunks = max(round(table.nbytes / DEFAULT_ARROW_CHUNK_BYTES_SIZE), 1)

    # Clamp to the maximum number of chunks
    num_chunks = min(num_chunks, DEFAULT_MAX_NUM_CHUNKS)

    return math.ceil(table.num_rows / num_chunks)


def validate_accessor_length_matches_table(
    accessor: Array | ChunkedArray,
    table: Table,
) -> None:
    if len(accessor) != len(table):
        raise TraitError("accessor must have same length as table")


def serialize_view_state(data: ViewState | None, obj: Any) -> None | dict[str, Any]:  # noqa: ARG001
    if data is None:
        return None

    return data._asdict()


def serialize_timestamp_accessor(
    timestamps: ChunkedArray,
    obj: TripsLayer,
) -> list[bytes]:
    """Subtract off min timestamp to fit into f32 integer range.

    Then cast to float32.
    """
    # Note: this has some overlap with `timestamp_max_physical_value` in utils.
    # Cast to int64 type
    timestamps = timestamps.cast(DataType.list(DataType.int64()))

    start_offset_adjustment = Scalar(
        timestamp_start_offset(timestamps),
        type=DataType.int64(),
    )

    list_offsets_iter = list_offsets(timestamps)
    inner_values_iter = list_flatten(timestamps)

    offsetted_chunks = []
    for offsets, inner_values in zip(
        list_offsets_iter,
        inner_values_iter,
        strict=True,
    ):
        offsetted_values = ac.add(inner_values, start_offset_adjustment)
        f32_values = offsetted_values.cast(DataType.int64()).cast(DataType.float32())
        offsetted_chunks.append(list_array(offsets, f32_values))

    f32_timestamps_col = ChunkedArray(offsetted_chunks)
    return serialize_accessor(f32_timestamps_col, obj)


ACCESSOR_SERIALIZATION = {"to_json": serialize_accessor}
TIMESTAMP_ACCESSOR_SERIALIZATION = {"to_json": serialize_timestamp_accessor}
TABLE_SERIALIZATION = {"to_json": serialize_table}



from typing import NamedTuple


class ViewState(NamedTuple):
    """State of a view position of a map."""

    longitude: float
    """Longitude at the map center"""

    latitude: float
    """Latitude at the map center."""

    zoom: float
    """Zoom level."""

    pitch: float
    """Pitch angle in degrees. `0` is top-down."""

    bearing: float
    """Bearing angle in degrees. `0` is north."""



"""Colormap helpers."""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, TypeAlias

import numpy as np
from arro3.compute import dictionary_encode
from arro3.core import (
    Array,
    ChunkedArray,
    DataType,
    dictionary_dictionary,
    dictionary_indices,
)

from lonboard._vendor.matplotlib.colors import _to_rgba_no_colorcycle

if TYPE_CHECKING:
    import matplotlib as mpl
    import pandas as pd
    import pyarrow as pa
    from arro3.core.types import ArrowArrayExportable, ArrowStreamExportable
    from numpy.typing import NDArray
    from palettable.palette import Palette


__all__ = (
    "apply_categorical_cmap",
    "apply_continuous_cmap",
)

RGBColor: TypeAlias = (
    tuple[int, int, int] | tuple[int, int, int, int] | Sequence[int] | str
)
"""A type definition for an RGB or RGBA color value

All values must range between 0 and 255 (inclusive). If only three values are provided,
the fourth (alpha) channel will be inferred as 255 (meaning full opacity, no
transparency).

This can also be a string that refers to a [matplotlib named color](https://matplotlib.org/stable/gallery/color/named_colors.html).
"""

DiscreteColormap = Mapping[Any, RGBColor]
"""A type definition for a discrete colormap.

For example, for a land cover colormap, you may want to use the following dict:

```py
color_map = {
    "Open Water": [70, 107, 159],
    "Perennial Snow/Ice": [209, 222, 248],
    "Developed, Open Space": [222, 197, 197],
    "Developed, Low Intensity": [217, 146, 130],
    "Developed, Medium Intensity": [235, 0, 0],
    "Developed, High Intensity": [171, 0, 0],
    "Barren Land": [179, 172, 159],
    "Deciduous Forest": [104, 171, 95],
    "Evergreen Forest": [28, 95, 44],
    "Mixed Forest": [181, 197, 143],
    "Shrub/Scrub": [204, 184, 121],
    "Herbaceous": [223, 223, 194],
    "Hay/Pasture": [220, 217, 57],
    "Cultivated Crops": [171, 108, 40],
    "Woody Wetlands": [184, 217, 235],
    "Emergent Herbaceous Wetlands": [108, 159, 184],
}
```

This corresponds to the following well-known color palette from the [National Land Cover
Database](https://www.mrlc.gov/data/legends/national-land-cover-database-class-legend-and-description).

<a href="https://grasswiki.osgeo.org/wiki/NLCD_Land_Cover">
    <img
        src="https://grasswiki.osgeo.org/w/images/NLCD_Colour_Classification_Update.jpg"
        alt="NLCD Land Cover Legend"
        width="300px"
    />
</a>

Keep in mind that the type of the key of the `color_map` dictionary is important. If
your data is represented as a string, the keys of the `color_map` must also be
represented as strings. If your data is represented as an integer, the keys of the
colormap should be integers.
"""


def apply_continuous_cmap(
    values: NDArray[np.floating],
    cmap: Palette | mpl.colors.Colormap,
    *,
    alpha: float | NDArray[np.floating] | None = None,
) -> NDArray[np.uint8]:
    """Apply a colormap to a column of continuous values.

    If you pass in a `Palette` object from the
    [`palettable`](https://github.com/jiffyclub/palettable) package, this will use
    matplotlib's [LinearSegmentedColormap][matplotlib.colors.LinearSegmentedColormap]
    under the hood. As described in Matplotlib's referenced docstring:

    > The lookup table is generated using linear interpolation for each primary color,
    > with the 0-1 domain divided into any number of segments.

    This means that input values are linearly combined from the two nearest colormap
    colors.

    If you pass in a Matplotlib `Colormap` object, this will use the lookup method
    defined in that class.

    Args:
        values: A numpy array of floating point values ranging from 0 to 1.
        cmap: Any matplotlib `Colormap` or `Palette` object from the
            [`palettable`](https://github.com/jiffyclub/palettable) package.

    Keyword Args:
        alpha: Alpha must be a scalar between 0 and 1, a sequence of such floats with
            shape matching `values`, or None.

    Returns:
        A two dimensional numpy array with data type [np.uint8][numpy.uint8]. The second
            dimension will have a length of either `3` if `alpha` is `None`, or `4` is
            each color has an alpha value.

    """
    try:
        from palettable.palette import Palette
    except ImportError:
        Palette = None  # noqa: N806

    try:
        import matplotlib as mpl
    except ImportError:
        mpl = None

    if Palette is not None and isinstance(cmap, Palette):
        colors: NDArray[np.uint8] = cmap.mpl_colormap(values, alpha=alpha, bytes=True)  # type: ignore
    elif mpl is not None and isinstance(cmap, mpl.colors.Colormap):
        colors: NDArray[np.uint8] = cmap(values, alpha=alpha, bytes=True)  # type: ignore
    else:
        raise TypeError("Expected cmap to be a palettable or matplotlib colormap.")

    # If the alpha values are all 255, don't serialize
    if (colors[:, 3] == 255).all():
        return colors[:, :3]

    return colors


def apply_categorical_cmap(  # noqa: C901
    values: (
        NDArray
        | pd.Series
        | pa.Array
        | pa.ChunkedArray
        | ArrowArrayExportable
        | ArrowStreamExportable
    ),
    cmap: DiscreteColormap,
    *,
    alpha: float | None = None,
) -> NDArray[np.uint8]:
    """Apply a colormap to a column of categorical values.

    If you're working with categorical data in Pandas or GeoPandas, and **especially**
    when those categories are strings, you should use the [pandas Categorical
    type](https://pandas.pydata.org/docs/user_guide/categorical.html). This
    will use much less memory and be faster to operate on. `apply_categorical_cmap` will
    be 2-3x faster (and use less memory itself) when your data is already represented as
    a categorical data type.

    The key of the colormap must be the same as the data type of the column of values
    you pass in. _Usually_ this will be string, when you perform the lookup on a string
    column of data.

    Args:
        values: A numpy array, pandas Series, pyarrow Array or pyarrow ChunkedArray of
            data. The data type of this column must match the keys of the colormap.
        cmap: A dictionary mapping keys to colors. See [DiscreteColormap] for more
            information.

    Keyword Args:
        alpha: The _default_ alpha value for entries in the colormap that do not have an
            alpha value defined. Alpha must be an integer between 0 and 255 (inclusive).

    Returns:
        A two dimensional numpy array with data type [np.uint8][numpy.uint8]. The second
            dimension will have a length of either `3` if `alpha` is `None`, or `4` is
            each color has an alpha value.

    """
    if isinstance(values, np.ndarray):
        values = Array.from_numpy(values)

    try:
        import pandas as pd

        if isinstance(values, pd.Series):
            values = Array.from_numpy(values)
    except ImportError:
        pass

    values = ChunkedArray(values)

    if not DataType.is_dictionary(values.type):
        values = ChunkedArray(dictionary_encode(values))

    dictionary = ChunkedArray(dictionary_dictionary(values))
    indices = ChunkedArray(dictionary_indices(values))

    # Build lookup table
    lut = np.zeros((len(dictionary), 4), dtype=np.uint8)
    if alpha is not None:
        assert isinstance(alpha, int), "alpha must be an integer"
        assert 0 <= alpha <= 255, "alpha must be between 0-255 (inclusive)."

        lut[:, 3] = alpha
    else:
        lut[:, 3] = 255

    for i, key in enumerate(dictionary):
        color = cmap[key.as_py()]

        if isinstance(color, str):
            color = _to_rgba_no_colorcycle(color, alpha=alpha)
            color = [c * 255 for c in color]

        if len(color) == 3:
            lut[i, :3] = color
        elif len(color) == 4:
            lut[i] = color
        else:
            raise ValueError(
                "Expected color to be 3 or 4 values representing RGB or RGBA.",
            )

    colors = lut[indices]

    # If the alpha values are all 255, don't serialize
    if (colors[:, 3] == 255).all():
        return colors[:, :3]

    return colors



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 12):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

if TYPE_CHECKING:
    from lonboard.basemap import CartoBasemap


class MapKwargs(TypedDict, total=False):
    """Kwargs to pass into map constructor."""

    height: int | str
    basemap_style: str | CartoBasemap
    parameters: dict[str, Any]
    picking_radius: int
    show_tooltip: bool
    show_side_panel: bool
    use_device_pixels: int | float | bool
    view_state: dict[str, Any]
